# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members

import queue
import threading

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import log

from buildbot.util import backoff


class UndoableQueue(queue.Queue):
    def unget(self, x):
        with self.mutex:
            self.queue.appendleft(x)


class _TerminateRequest:
    pass


class ConnectableThreadQueue(threading.Thread):
    """
    This provides worker thread that is processing work given via execute_in_thread() method.
    The return value of the function submitted to execute_in_thread() is returned via Deferred.
    All work is performed in a "connection", which is established in create_connection() which
    is intended to be overridden by user. The user is expected to return an opaque connection
    object from create_connection(). create_connection() must not throw exceptions.
    The connection is from the user-side closed by calling close_connection().

    The connection is passed as the first argument to the functions submitted to
    execute_in_thread().

    When the thread is joined, it will execute all currently pending items and call
    on_close_connection() if needed to close the connection. Any work submitted after join() is
    called will be ignored.
    """

    def __init__(
        self,
        connect_backoff_start_seconds=1,
        connect_backoff_multiplier=1.1,
        connect_backoff_max_wait_seconds=3600,
    ):
        self._queue = UndoableQueue()
        self._conn = None
        self._backoff_engine = backoff.ExponentialBackoffEngineSync(
            start_seconds=connect_backoff_start_seconds,
            multiplier=connect_backoff_multiplier,
            max_wait_seconds=connect_backoff_max_wait_seconds,
        )

        super().__init__(daemon=True)
        self.connecting = False
        self.start()

    def join(self, *args, **kwargs):
        self.execute_in_thread(_TerminateRequest())
        super().join(*args, **kwargs)

    def execute_in_thread(self, cb, *args, **kwargs):
        d = defer.Deferred()
        self._queue.put((d, cb, args, kwargs))
        return d

    @property
    def conn(self):
        return self._conn

    def close_connection(self):
        self._conn = None
        self.connecting = False

    def on_close_connection(self, conn):
        # override to perform any additional connection closing tasks
        self.close_connection()

    def create_connection(self):
        # override to create a new connection
        raise NotImplementedError()

    def _handle_backoff(self, msg):
        # returns True if termination has been requested
        log.err(msg)
        try:
            self._backoff_engine.wait_on_failure()
        except backoff.BackoffTimeoutExceededError:
            self._backoff_engine.on_success()  # reset the timers
            if self._drain_queue_with_exception(backoff.BackoffTimeoutExceededError(msg)):
                return True
        return False

    def _drain_queue_with_exception(self, e):
        # returns True if termination has been requested
        try:
            while True:
                result_d, next_operation, _, __ = self._queue.get(block=False)
                if isinstance(next_operation, _TerminateRequest):
                    self._queue.task_done()
                    reactor.callFromThread(result_d.callback, None)
                    return True
                else:
                    self._queue.task_done()
                    reactor.callFromThread(result_d.errback, e)
        except queue.Empty:
            return False

    def run(self):
        while True:
            result_d, next_operation, args, kwargs = self._queue.get()

            if isinstance(next_operation, _TerminateRequest):
                self._queue.task_done()
                reactor.callFromThread(result_d.callback, None)
                break

            if not self._conn:
                self.connecting = True
                self._queue.unget((result_d, next_operation, args, kwargs))
                try:
                    self._conn = self.create_connection()
                    self.connecting = False

                    if self._conn is not None:
                        self._backoff_engine.on_success()
                    elif self._handle_backoff('Did not receive connection'):
                        break

                except Exception as e:
                    self.connecting = False
                    if self._handle_backoff(f'Exception received: {e}'):
                        break
                continue
            try:
                result = next_operation(self._conn, *args, **kwargs)
                reactor.callFromThread(result_d.callback, result)
            except Exception as e:
                reactor.callFromThread(result_d.errback, e)
            self._queue.task_done()

        if self._conn is not None:
            self.on_close_connection(self._conn)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from __future__ import annotations

import os
import queue
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable

import sqlalchemy as sa
from twisted.internet import defer

from buildbot import config as config_module
from buildbot.db import connector
from buildbot.db import exceptions
from buildbot.db import model
from buildbot.master import BuildMaster
from buildbot.scripts import base
from buildbot.util import in_reactor
from buildbot.util import misc

if TYPE_CHECKING:
    from buildbot.util.twisted import InlineCallbacksType


@in_reactor
def copy_database(config: dict) -> defer.Deferred:  # pragma: no cover
    # we separate the actual implementation to protect unit tests
    # from @in_reactor which stops the reactor
    return _copy_database_in_reactor(config)


@defer.inlineCallbacks
def _copy_database_in_reactor(config: dict) -> InlineCallbacksType[int]:
    if not base.checkBasedir(config):
        return 1

    print_debug = not config["quiet"]
    ignore_fk_error_rows = config['ignore-fk-error-rows']

    def print_log(*args: Any, **kwargs: Any) -> None:
        if print_debug:
            print(*args, **kwargs)

    config['basedir'] = os.path.abspath(config['basedir'])

    with base.captureErrors(
        (SyntaxError, ImportError), f"Unable to load 'buildbot.tac' from '{config['basedir']}':"
    ):
        config_file = base.getConfigFileFromTac(config['basedir'])

    if not config_file:
        return 1

    with base.captureErrors(
        config_module.ConfigErrors, f"Unable to load '{config_file}' from '{config['basedir']}':"
    ):
        master_src_cfg = base.loadConfig(config, config_file)
        master_dst_cfg = base.loadConfig(config, config_file)

    if not master_src_cfg or not master_dst_cfg:
        return 1

    master_dst_cfg.db.db_url = config["destination_url"]

    print_log(f"Copying database ({master_src_cfg.db.db_url}) to ({config['destination_url']})")

    if not master_src_cfg or not master_dst_cfg:
        return 1

    master_src = None
    master_dst = None
    try:
        master_src = BuildMaster(config['basedir'])
        master_src.config = master_src_cfg
        try:
            yield master_src.db.setup(check_version=True, verbose=not config["quiet"])
        except exceptions.DatabaseNotReadyError:
            for l in connector.upgrade_message.format(basedir=config['basedir']).split('\n'):
                print(l)
            return 1

        master_dst = BuildMaster(config['basedir'])
        master_dst.config = master_dst_cfg
        yield master_dst.db.setup(check_version=False, verbose=not config["quiet"])
        yield master_dst.db.model.upgrade()
        yield _copy_database_with_db(master_src.db, master_dst.db, ignore_fk_error_rows, print_log)
    finally:
        for master in (master_src, master_dst):
            if master is not None and master.db.pool is not None:
                yield master.db.pool.stop()

    return 0


def _thd_check_row_foreign_keys(
    table_name: str,
    row_dict: dict[str, Any],
    column_name: str,
    id_rows: set[Any],
    print_log: Callable[..., None],
) -> bool:
    if column_name not in row_dict:
        return True

    value = row_dict[column_name]
    if value is None:
        return True

    if value not in id_rows:
        row_str = repr(row_dict)[0:200]
        print_log(
            f'Ignoring row from {table_name} because {column_name}={value} foreign key '
            f'constraint failed. Row: {row_str}'
        )
        return False

    return True


def _thd_check_rows_foreign_keys(
    table_name: str,
    row_dicts: list[dict[str, Any]],
    column_name: str,
    id_rows: set[Any],
    print_log: Callable[..., None],
) -> list[dict[str, Any]]:
    return [
        row_dict
        for row_dict in row_dicts
        if _thd_check_row_foreign_keys(table_name, row_dict, column_name, id_rows, print_log)
    ]


@defer.inlineCallbacks
def _copy_single_table(
    metadata: Any,
    src_db: Any,
    dst_db: Any,
    table_name: str,
    buildset_to_parent_buildid: list[tuple[int, int]],
    buildset_to_rebuilt_buildid: list[tuple[int, int]],
    ignore_fk_error_rows: bool,
    print_log: Callable[..., None],
) -> InlineCallbacksType[None]:
    table = metadata.tables[table_name]
    column_keys = table.columns.keys()

    rows_queue: queue.Queue[Any] = queue.Queue(32)
    written_count = [0]
    total_count = [0]

    autoincrement_foreign_key_column = None
    foreign_key_check_columns = []

    for column_name, column in table.columns.items():
        if not column.foreign_keys and column.primary_key and isinstance(column.type, sa.Integer):
            autoincrement_foreign_key_column = column_name

        for fk in column.foreign_keys:
            if table_name == 'buildsets' and column_name in ('parent_buildid', 'rebuilt_buildid'):
                continue
            if table_name == 'changes' and column_name in ('parent_changeids',):
                # TODO: not currently handled because column refers to the same table
                continue
            foreign_key_check_columns.append((column_name, fk.column))

    def tdh_query_all_column_rows(conn: sa.engine.Connection, column: sa.Column) -> set[Any]:
        q = sa.select(column).select_from(column.table)
        result = conn.execute(q)

        # Load data incrementally in order to control maximum used memory size
        ids = set()
        while True:
            chunk = result.fetchmany(10000)
            if not chunk:
                break
            for row in chunk:
                ids.add(getattr(row, column.name))
        return ids

    got_error = False

    def thd_write(conn: sa.engine.Connection) -> None:
        max_column_id = 0

        foreign_key_check_rows = []
        if ignore_fk_error_rows:
            foreign_key_check_rows = [
                (column_name, tdh_query_all_column_rows(conn, fk_column))
                for column_name, fk_column in foreign_key_check_columns
            ]

        while True:
            try:
                rows = rows_queue.get(timeout=1)
                if rows is None:
                    if autoincrement_foreign_key_column is not None and max_column_id != 0:
                        if dst_db.pool.engine.dialect.name == 'postgresql':
                            # Explicitly inserting primary row IDs does not bump the primary key
                            # sequence on Postgres
                            seq_name = f"{table_name}_{autoincrement_foreign_key_column}_seq"
                            transaction = conn.begin()
                            conn.execute(
                                sa.text(
                                    f"ALTER SEQUENCE {seq_name} RESTART WITH {max_column_id + 1}"
                                )
                            )
                            transaction.commit()

                    rows_queue.task_done()
                    return

                row_dicts = [{k: getattr(row, k) for k in column_keys} for row in rows]

                if autoincrement_foreign_key_column is not None:
                    for row in row_dicts:
                        max_column_id = max(max_column_id, row[autoincrement_foreign_key_column])

                if ignore_fk_error_rows:
                    for column_name, id_rows in foreign_key_check_rows:
                        row_dicts = _thd_check_rows_foreign_keys(
                            table_name, row_dicts, column_name, id_rows, print_log
                        )

                if table_name == "buildsets":
                    for row_dict in row_dicts:
                        if row_dict["parent_buildid"] is not None:
                            buildset_to_parent_buildid.append((
                                row_dict["id"],
                                row_dict["parent_buildid"],
                            ))
                        row_dict["parent_buildid"] = None

                        if row_dict['rebuilt_buildid'] is not None:
                            buildset_to_rebuilt_buildid.append((
                                row_dict['id'],
                                row_dict['rebuilt_buildid'],
                            ))
                        row_dict['rebuilt_buildid'] = None

            except queue.Empty:
                continue
            except Exception:
                nonlocal got_error
                got_error = True
                # unblock queue
                try:
                    rows_queue.get(timeout=1)
                    rows_queue.task_done()
                except queue.Empty:
                    pass
                raise

            try:
                written_count[0] += len(rows)
                print_log(
                    f"Copying {len(rows)} items ({written_count[0]}/{total_count[0]}) "
                    f"for {table_name} table"
                )

                if len(row_dicts) > 0:
                    conn.execute(table.insert(), row_dicts)
                    conn.commit()

            finally:
                rows_queue.task_done()

    def thd_read(conn: sa.engine.Connection) -> None:
        q = sa.select(sa.sql.func.count()).select_from(table)
        total_count[0] = conn.execute(q).scalar() or 0

        result = conn.execute(sa.select(table))
        while not got_error:
            chunk = result.fetchmany(10000)
            if not chunk:
                break
            rows_queue.put(chunk)

        rows_queue.put(None)

    error: Exception | None = None
    tasks = [src_db.pool.do(thd_read), dst_db.pool.do(thd_write)]
    for d in tasks:
        try:
            yield d
        except Exception as e:
            error = e

    rows_queue.join()

    if error is not None:
        raise error


@defer.inlineCallbacks
def _copy_database_with_db(
    src_db: Any, dst_db: Any, ignore_fk_error_rows: bool, print_log: Callable[..., None]
) -> InlineCallbacksType[None]:
    # Tables need to be specified in correct order so that tables that other tables depend on are
    # copied first.
    table_names = [
        # Note that buildsets.parent_buildid and rebuilt_buildid introduce circular dependency.
        # They are handled separately
        "buildsets",
        "buildset_properties",
        "projects",
        "codebases",
        "codebase_commits",
        "codebase_branches",
        "builders",
        "changesources",
        "buildrequests",
        "workers",
        "masters",
        "buildrequest_claims",
        "changesource_masters",
        "builder_masters",
        "configured_workers",
        "connected_workers",
        "patches",
        "sourcestamps",
        "buildset_sourcestamps",
        "changes",
        "change_files",
        "change_properties",
        "users",
        "users_info",
        "change_users",
        "builds",
        "build_properties",
        "build_data",
        "steps",
        "logs",
        "logchunks",
        "schedulers",
        "scheduler_masters",
        "scheduler_changes",
        "tags",
        "builders_tags",
        "test_result_sets",
        "test_names",
        "test_code_paths",
        "test_results",
        "objects",
        "object_state",
    ]

    metadata = model.Model.metadata
    assert len(set(table_names)) == len(set(metadata.tables.keys()))

    # Not a dict so that the values are inserted back in predictable order
    buildset_to_parent_buildid: list[tuple[int, int]] = []
    buildset_to_rebuilt_buildid: list[tuple[int, int]] = []

    for table_name in table_names:
        yield _copy_single_table(
            metadata,
            src_db,
            dst_db,
            table_name,
            buildset_to_parent_buildid,
            buildset_to_rebuilt_buildid,
            ignore_fk_error_rows,
            print_log,
        )

    def thd_write_buildset_parent_buildid(conn: sa.engine.Connection) -> None:
        written_count = 0
        for rows in misc.chunkify_list(buildset_to_parent_buildid, 10000):
            q = model.Model.buildsets.update()
            q = q.where(model.Model.buildsets.c.id == sa.bindparam('_id'))
            q = q.values({'parent_buildid': sa.bindparam('parent_buildid')})

            written_count += len(rows)
            print_log(
                f"Copying {len(rows)} items ({written_count}/{len(buildset_to_parent_buildid)}) "
                f"for buildset.parent_buildid field"
            )

            conn.execute(
                q,
                [
                    {'_id': buildset_id, 'parent_buildid': parent_buildid}
                    for buildset_id, parent_buildid in rows
                ],
            )

    yield dst_db.pool.do(thd_write_buildset_parent_buildid)

    def thd_write_buildset_rebuilt_buildid(conn: sa.engine.Connection) -> None:
        written_count = 0
        for rows in misc.chunkify_list(buildset_to_rebuilt_buildid, 10000):
            q = model.Model.buildsets.update()
            q = q.where(model.Model.buildsets.c.id == sa.bindparam('_id'))
            q = q.values({'rebuilt_buildid': sa.bindparam('rebuilt_buildid')})

            written_count += len(rows)
            print_log(
                f"Copying {len(rows)} items ({written_count}/{len(buildset_to_rebuilt_buildid)}) "
                f"for buildset.rebuilt_buildid field"
            )

            conn.execute(
                q,
                [
                    {'_id': buildset_id, 'rebuilt_buildid': rebuilt_buildid}
                    for buildset_id, rebuilt_buildid in rows
                ],
            )

    yield dst_db.pool.do(thd_write_buildset_rebuilt_buildid)

    print_log("Copy complete")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import os
import tempfile

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.config import BuilderConfig
from buildbot.plugins import changes
from buildbot.plugins import schedulers
from buildbot.process import factory
from buildbot.steps.configurable import BuildbotCiSetupSteps
from buildbot.steps.configurable import BuildbotTestCiTrigger
from buildbot.steps.source.git import Git
from buildbot.test.util.git_repository import TestGitRepository
from buildbot.test.util.integration import RunMasterBase
from buildbot.worker import Worker

buildbot_ci_yml = """\
language: python

label_mapping:
  TWISTED: tw
  SQLALCHEMY: sqla
  SQLALCHEMY_MIGRATE: sqlam
  latest: l
  python: py

python:
  - "3.9"
env:
  global:
      - CI=true
matrix:
  include:
    # Test different versions of SQLAlchemy
    - python: "3.9"
      env: TWISTED=12.0.0 SQLALCHEMY=0.6.0 SQLALCHEMY_MIGRATE=0.7.1
    - python: "3.9"
      env: TWISTED=13.0.0 SQLALCHEMY=0.6.8 SQLALCHEMY_MIGRATE=0.7.1
    - python: "3.9"
      env: TWISTED=14.0.0 SQLALCHEMY=0.6.8 SQLALCHEMY_MIGRATE=0.7.1
    - python: "3.9"
      env: TWISTED=15.0.0 SQLALCHEMY=0.6.8 SQLALCHEMY_MIGRATE=0.7.1

before_install:
  - echo doing before install
  - echo doing before install 2nd command
install:
  - echo doing install
script:
  - echo doing scripts
  - echo Ran 10 tests with 0 failures and 0 errors
after_success:
  - echo doing after success
notifications:
  email: false

"""


class BuildbotTestCiTest(RunMasterBase):
    timeout = 300

    @defer.inlineCallbacks
    def setUp(self):
        yield super().setUp()
        try:
            self.repo = TestGitRepository(
                repository_path=tempfile.mkdtemp(
                    prefix="TestRepository_",
                    dir=os.getcwd(),
                )
            )
        except FileNotFoundError as e:
            raise unittest.SkipTest("Can't find git binary") from e

        self.prepare_repository()

    def prepare_repository(self):
        self.repo.create_file_text('.bbtravis.yml', buildbot_ci_yml)
        self.repo.exec_git(['add', '.bbtravis.yml'])
        self.repo.commit(message='Initial commit', files=['.bbtravis.yml'])

    @defer.inlineCallbacks
    def setup_config(self):
        c = {
            'workers': [Worker("local1", "p")],
            'services': [
                changes.GitPoller(
                    repourl=str(self.repo.repository_path), project='buildbot', branch='main'
                )
            ],
            'builders': [],
            'schedulers': [],
            'multiMaster': True,
        }

        repository = str(self.repo.repository_path)
        job_name = "buildbot-job"
        try_name = "buildbot-try"
        spawner_name = "buildbot"

        codebases = {spawner_name: {'repository': repository}}

        # main job
        f = factory.BuildFactory()
        f.addStep(Git(repourl=repository, codebase=spawner_name, mode='incremental'))
        f.addStep(BuildbotCiSetupSteps())

        c['builders'].append(
            BuilderConfig(
                name=job_name,
                workernames=['local1'],
                collapseRequests=False,
                tags=["job", 'buildbot'],
                factory=f,
            )
        )

        c['schedulers'].append(
            schedulers.Triggerable(name=job_name, builderNames=[job_name], codebases=codebases)
        )

        # spawner
        f = factory.BuildFactory()
        f.addStep(Git(repourl=repository, codebase=spawner_name, mode='incremental'))
        f.addStep(BuildbotTestCiTrigger(scheduler=job_name))
        c['builders'].append(
            BuilderConfig(
                name=spawner_name,
                workernames=['local1'],
                properties={'BUILDBOT_PULL_REQUEST': True},
                tags=["spawner", 'buildbot'],
                factory=f,
            )
        )

        c['schedulers'].append(
            schedulers.AnyBranchScheduler(
                name=spawner_name, builderNames=[spawner_name], codebases=codebases
            )
        )

        # try job
        f = factory.BuildFactory()
        f.addStep(Git(repourl=repository, codebase=spawner_name, mode='incremental'))
        f.addStep(BuildbotTestCiTrigger(scheduler=job_name))

        c['builders'].append(
            BuilderConfig(
                name=try_name,
                workernames=['local1'],
                properties={'BUILDBOT_PULL_REQUEST': True},
                tags=["try", 'buildbot'],
                factory=f,
            )
        )

        yield self.setup_master(c)

    @defer.inlineCallbacks
    def test_buildbot_ci(self):
        yield self.setup_config()
        change = dict(
            branch="main",
            files=["foo.c"],
            author="me@foo.com",
            comments="good stuff",
            revision="HEAD",
            repository=str(self.repo.repository_path),
            codebase='buildbot',
            project="buildbot",
        )
        build = yield self.doForceBuild(wantSteps=True, useChange=change, wantLogs=True)
        if 'worker local1 ready' in build['steps'][0]['state_string']:
            build['steps'] = build['steps'][1:]
        self.assertEqual(build['steps'][0]['state_string'], 'update buildbot')
        self.assertEqual(build['steps'][0]['name'], 'git-buildbot')
        self.assertEqual(
            build['steps'][1]['state_string'],
            'triggered ' + ", ".join(["buildbot-job"] * 4) + ', 4 successes',
        )
        url_names = [url['name'] for url in build['steps'][1]['urls']]
        url_urls = [url['url'] for url in build['steps'][1]['urls']]
        self.assertIn('http://localhost:8080/#/builders/4/builds/1', url_urls)
        self.assertIn('success: buildbot py:3.9 sqla:0.6.0 sqlam:0.7.1 tw:12.0.0 #1', url_names)
        self.assertEqual(build['steps'][1]['logs'][0]['contents']['content'], buildbot_ci_yml)

        builds = yield self.master.data.get(("builds",))
        self.assertEqual(len(builds), 5)
        props = {}
        buildernames = {}
        labels = {}
        for build in builds:
            build['properties'] = yield self.master.data.get((
                "builds",
                build['buildid'],
                'properties',
            ))
            props[build['buildid']] = {
                k: v[0] for k, v in build['properties'].items() if v[1] == 'BuildbotTestCiTrigger'
            }
            buildernames[build['buildid']] = build['properties'].get('virtual_builder_name')
            labels[build['buildid']] = build['properties'].get('matrix_label')
        self.assertEqual(
            props,
            {
                1: {},
                2: {
                    'python': '3.9',
                    'CI': 'true',
                    'TWISTED': '12.0.0',
                    'SQLALCHEMY': '0.6.0',
                    'SQLALCHEMY_MIGRATE': '0.7.1',
                    'virtual_builder_name': 'buildbot py:3.9 sqla:0.6.0 sqlam:0.7.1 tw:12.0.0',
                    'virtual_builder_tags': [
                        'buildbot',
                        'py:3.9',
                        'sqla:0.6.0',
                        'sqlam:0.7.1',
                        'tw:12.0.0',
                        '_virtual_',
                    ],
                    'matrix_label': 'py:3.9/sqla:0.6.0/sqlam:0.7.1/tw:12.0.0',
                },
                3: {
                    'python': '3.9',
                    'CI': 'true',
                    'TWISTED': '13.0.0',
                    'SQLALCHEMY': '0.6.8',
                    'SQLALCHEMY_MIGRATE': '0.7.1',
                    'virtual_builder_name': 'buildbot py:3.9 sqla:0.6.8 sqlam:0.7.1 tw:13.0.0',
                    'virtual_builder_tags': [
                        'buildbot',
                        'py:3.9',
                        'sqla:0.6.8',
                        'sqlam:0.7.1',
                        'tw:13.0.0',
                        '_virtual_',
                    ],
                    'matrix_label': 'py:3.9/sqla:0.6.8/sqlam:0.7.1/tw:13.0.0',
                },
                4: {
                    'python': '3.9',
                    'CI': 'true',
                    'TWISTED': '14.0.0',
                    'SQLALCHEMY': '0.6.8',
                    'SQLALCHEMY_MIGRATE': '0.7.1',
                    'virtual_builder_name': 'buildbot py:3.9 sqla:0.6.8 sqlam:0.7.1 tw:14.0.0',
                    'virtual_builder_tags': [
                        'buildbot',
                        'py:3.9',
                        'sqla:0.6.8',
                        'sqlam:0.7.1',
                        'tw:14.0.0',
                        '_virtual_',
                    ],
                    'matrix_label': 'py:3.9/sqla:0.6.8/sqlam:0.7.1/tw:14.0.0',
                },
                5: {
                    'python': '3.9',
                    'CI': 'true',
                    'TWISTED': '15.0.0',
                    'SQLALCHEMY': '0.6.8',
                    'SQLALCHEMY_MIGRATE': '0.7.1',
                    'virtual_builder_name': 'buildbot py:3.9 sqla:0.6.8 sqlam:0.7.1 tw:15.0.0',
                    'virtual_builder_tags': [
                        'buildbot',
                        'py:3.9',
                        'sqla:0.6.8',
                        'sqlam:0.7.1',
                        'tw:15.0.0',
                        '_virtual_',
                    ],
                    'matrix_label': 'py:3.9/sqla:0.6.8/sqlam:0.7.1/tw:15.0.0',
                },
            },
        )
        # global env CI should not be there
        self.assertEqual(
            buildernames,
            {
                1: None,
                2: ('buildbot py:3.9 sqla:0.6.0 sqlam:0.7.1 tw:12.0.0', 'BuildbotTestCiTrigger'),
                3: ('buildbot py:3.9 sqla:0.6.8 sqlam:0.7.1 tw:13.0.0', 'BuildbotTestCiTrigger'),
                4: ('buildbot py:3.9 sqla:0.6.8 sqlam:0.7.1 tw:14.0.0', 'BuildbotTestCiTrigger'),
                5: ('buildbot py:3.9 sqla:0.6.8 sqlam:0.7.1 tw:15.0.0', 'BuildbotTestCiTrigger'),
            },
        )
        self.assertEqual(
            labels,
            {
                1: None,
                2: ('py:3.9/sqla:0.6.0/sqlam:0.7.1/tw:12.0.0', 'BuildbotTestCiTrigger'),
                3: ('py:3.9/sqla:0.6.8/sqlam:0.7.1/tw:13.0.0', 'BuildbotTestCiTrigger'),
                4: ('py:3.9/sqla:0.6.8/sqlam:0.7.1/tw:14.0.0', 'BuildbotTestCiTrigger'),
                5: ('py:3.9/sqla:0.6.8/sqlam:0.7.1/tw:15.0.0', 'BuildbotTestCiTrigger'),
            },
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from twisted.python import log
from twisted.trial import unittest

from buildbot.util import lineboundaries


class TestLineBoundaryFinder(unittest.TestCase):
    def setUp(self):
        self.lbf = lineboundaries.LineBoundaryFinder()

    def test_already_terminated(self):
        res = self.lbf.append('abcd\ndefg\n')
        self.assertEqual(res, 'abcd\ndefg\n')
        res = self.lbf.append('xyz\n')
        self.assertEqual(res, 'xyz\n')
        res = self.lbf.flush()
        self.assertEqual(res, None)

    def test_partial_line(self):
        res = self.lbf.append('hello\nworld')
        self.assertEqual(res, 'hello\n')
        res = self.lbf.flush()
        self.assertEqual(res, 'world\n')

    def test_empty_appends(self):
        res = self.lbf.append('hello ')
        self.assertEqual(res, None)

        res = self.lbf.append('')
        self.assertEqual(res, None)

        res = self.lbf.append('world\n')
        self.assertEqual(res, 'hello world\n')

        res = self.lbf.append('')
        self.assertEqual(res, None)

    def test_embedded_newlines(self):
        res = self.lbf.append('hello, ')
        self.assertEqual(res, None)

        res = self.lbf.append('cruel\nworld')
        self.assertEqual(res, 'hello, cruel\n')

        res = self.lbf.flush()
        self.assertEqual(res, 'world\n')

    def test_windows_newlines_folded(self):
        r"Windows' \r\n is treated as and converted to a newline"
        res = self.lbf.append('hello, ')
        self.assertEqual(res, None)

        res = self.lbf.append('cruel\r\n\r\nworld')
        self.assertEqual(res, 'hello, cruel\n\n')

        res = self.lbf.flush()
        self.assertEqual(res, 'world\n')

    def test_bare_cr_folded(self):
        r"a bare \r is treated as and converted to a newline"
        self.lbf.append('1%\r5%\r15%\r100%\nfinished')
        res = self.lbf.flush()
        self.assertEqual(res, 'finished\n')

    def test_backspace_folded(self):
        r"a lot of \b is treated as and converted to a newline"
        self.lbf.append('1%\b\b5%\b\b15%\b\b\b100%\nfinished')
        res = self.lbf.flush()
        self.assertEqual(res, 'finished\n')

    def test_mixed_consecutive_newlines(self):
        r"mixing newline styles back-to-back doesn't collapse them"
        res = self.lbf.append('1\r\n\n\r')
        self.assertEqual(res, '1\n\n')

        res = self.lbf.append('2\n\r\n')
        self.assertEqual(res, '\n2\n\n')

    def test_split_newlines(self):
        r"multi-character newlines, split across chunks, are converted"
        input = 'a\nb\r\nc\rd\n\re'
        result = []
        for splitpoint in range(1, len(input) - 1):
            a, b = input[:splitpoint], input[splitpoint:]
            result.append(self.lbf.append(a))
            result.append(self.lbf.append(b))
            result.append(self.lbf.flush())

            result = [e for e in result if e is not None]
            res = ''.join(result)

            log.msg(f'feeding {a!r}, {b!r} gives {res!r}')
            self.assertEqual(res, 'a\nb\nc\nd\n\ne\n')
            result.clear()

    def test_split_terminal_control(self):
        """terminal control characters are converted"""
        res = self.lbf.append('1234\033[u4321')
        self.assertEqual(res, '1234\n')

        res = self.lbf.flush()
        self.assertEqual(res, '4321\n')

        res = self.lbf.append('1234\033[1;2H4321')
        self.assertEqual(res, '1234\n')

        res = self.lbf.flush()
        self.assertEqual(res, '4321\n')

        res = self.lbf.append('1234\033[1;2f4321')
        self.assertEqual(res, '1234\n')

        res = self.lbf.flush()
        self.assertEqual(res, '4321\n')

    def test_long_lines(self):
        """long lines are split"""
        res = []
        for _ in range(4):
            res.append(self.lbf.append('12' * 1000))
        res = [e for e in res if e is not None]
        res = ''.join(res)
        # a split at 4096 + the remaining chars
        self.assertEqual(res, '12' * 2048 + '\n' + '12' * 952 + '\n')

    def test_huge_lines(self):
        """huge lines are split"""
        res = []
        res.append(self.lbf.append('12' * 32768))
        res.append(self.lbf.flush())
        res = [e for e in res if e is not None]
        self.assertEqual(res, [('12' * 2048 + '\n') * 16])

    def test_empty_flush(self):
        res = self.lbf.flush()
        self.assertEqual(res, None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import base64

import msgpack
from autobahn.twisted.websocket import WebSocketClientFactory
from autobahn.twisted.websocket import WebSocketClientProtocol
from autobahn.websocket.types import ConnectingRequest
from twisted.internet import defer
from twisted.python import log

from buildbot_worker.base import ProtocolCommandBase
from buildbot_worker.util import deferwaiter


class RemoteWorkerError(Exception):
    pass


def decode_http_authorization_header(value):
    if value[:5] != 'Basic':
        raise ValueError("Value should always start with 'Basic'")

    credentials_str = base64.b64decode(value[6:]).decode()
    if ':' not in credentials_str:
        raise ValueError("String of credentials should always have a colon.")

    username, password = credentials_str.split(':', maxsplit=1)
    return (username, password)


def encode_http_authorization_header(name, password):
    if b":" in name:
        raise ValueError("Username is not allowed to contain a colon.")
    userpass = name + b':' + password
    return 'Basic ' + base64.b64encode(userpass).decode()


def remote_print(self, message):
    log.msg(f"WorkerForBuilder.remote_print({self.name}): message from master: {message}")


class ProtocolCommandMsgpack(ProtocolCommandBase):
    def __init__(
        self,
        unicode_encoding,
        worker_basedir,
        buffer_size,
        buffer_timeout,
        max_line_length,
        newline_re,
        builder_is_running,
        on_command_complete,
        protocol,
        command_id,
        command,
        args,
    ):
        ProtocolCommandBase.__init__(
            self,
            unicode_encoding,
            worker_basedir,
            buffer_size,
            buffer_timeout,
            max_line_length,
            newline_re,
            builder_is_running,
            on_command_complete,
            None,
            command,
            command_id,
            args,
        )
        self.protocol = protocol

    def protocol_args_setup(self, command, args):
        if "want_stdout" in args:
            if args["want_stdout"]:
                args["want_stdout"] = 1
            else:
                args["want_stdout"] = 0

        if "want_stderr" in args:
            if args["want_stderr"]:
                args["want_stderr"] = 1
            else:
                args["want_stderr"] = 0

        # to silence the ValueError in class Command() init
        if (command in ("upload_directory", "upload_file")) and 'writer' not in args:
            args['writer'] = None
        if command == "download_file" and 'reader' not in args:
            args['reader'] = None

    def protocol_send_update_message(self, message):
        d = self.protocol.get_message_result({
            'op': 'update',
            'args': message,
            'command_id': self.command_id,
        })
        d.addErrback(self._ack_failed, "ProtocolCommandBase.send_update")

    def protocol_notify_on_disconnect(self):
        pass

    @defer.inlineCallbacks
    def protocol_complete(self, failure):
        d_update = self.flush_command_output()
        if failure is not None:
            failure = str(failure)
        d_complete = self.protocol.get_message_result({
            'op': 'complete',
            'args': failure,
            'command_id': self.command_id,
        })
        yield d_update
        yield d_complete

    # Returns a Deferred
    def protocol_update_upload_file_close(self, writer):
        return self.protocol.get_message_result({
            'op': 'update_upload_file_close',
            'command_id': self.command_id,
        })

    # Returns a Deferred
    def protocol_update_upload_file_utime(self, writer, access_time, modified_time):
        return self.protocol.get_message_result({
            'op': 'update_upload_file_utime',
            'access_time': access_time,
            'modified_time': modified_time,
            'command_id': self.command_id,
        })

    # Returns a Deferred
    def protocol_update_upload_file_write(self, writer, data):
        return self.protocol.get_message_result({
            'op': 'update_upload_file_write',
            'args': data,
            'command_id': self.command_id,
        })

    # Returns a Deferred
    def protocol_update_upload_directory(self, writer):
        return self.protocol.get_message_result({
            'op': 'update_upload_directory_unpack',
            'command_id': self.command_id,
        })

    # Returns a Deferred
    def protocol_update_upload_directory_write(self, writer, data):
        return self.protocol.get_message_result({
            'op': 'update_upload_directory_write',
            'args': data,
            'command_id': self.command_id,
        })

    # Returns a Deferred
    def protocol_update_read_file_close(self, reader):
        return self.protocol.get_message_result({
            'op': 'update_read_file_close',
            'command_id': self.command_id,
        })

    # Returns a Deferred
    def protocol_update_read_file(self, reader, length):
        return self.protocol.get_message_result({
            'op': 'update_read_file',
            'length': length,
            'command_id': self.command_id,
        })


class ConnectionLostError(Exception):
    pass


class BuildbotWebSocketClientProtocol(WebSocketClientProtocol):
    debug = True

    def __init__(self):
        super().__init__()
        self.seq_num_to_waiters_map = {}
        self._deferwaiter = deferwaiter.DeferWaiter()

    def onConnect(self, response):
        if self.debug:
            log.msg(f"Server connected: {response.peer}")

    def onConnecting(self, transport_details):
        if self.debug:
            log.msg(f"Connecting; transport details: {transport_details}")

        auth_header = encode_http_authorization_header(self.factory.name, self.factory.password)

        return ConnectingRequest(
            host=self.factory.host,
            port=self.factory.port,
            resource=self.factory.resource,
            headers={"Authorization": auth_header},
            useragent=self.factory.useragent,
            origin=self.factory.origin,
            protocols=self.factory.protocols,
        )

    def maybe_log_worker_to_master_msg(self, message):
        if self.debug:
            log.msg("WORKER -> MASTER message: ", message)

    def maybe_log_master_to_worker_msg(self, message):
        if self.debug:
            log.msg("MASTER -> WORKER message: ", message)

    def contains_msg_key(self, msg, keys):
        for k in keys:
            if k not in msg:
                raise KeyError(f'message did not contain obligatory "{k}" key')

    def onOpen(self):
        if self.debug:
            log.msg("WebSocket connection open.")
        self.seq_number = 0

    def call_print(self, msg):
        is_exception = False
        try:
            self.contains_msg_key(msg, ('message',))
            self.factory.buildbot_bot.remote_print(msg['message'])
            result = None
        except Exception as e:
            is_exception = True
            result = str(e)

        self.send_response_msg(msg, result, is_exception)

    def call_keepalive(self, msg):
        result = None
        is_exception = False
        try:
            if self.debug:
                log.msg("Connection keepalive confirmed.")
        except Exception:
            pass

        self.send_response_msg(msg, result, is_exception)

    @defer.inlineCallbacks
    def call_get_worker_info(self, msg):
        is_exception = False
        try:
            result = yield self.factory.buildbot_bot.remote_getWorkerInfo()
        except Exception as e:
            is_exception = True
            result = str(e)
        self.send_response_msg(msg, result, is_exception)

    def call_set_worker_settings(self, msg):
        is_exception = False
        try:
            self.contains_msg_key(msg, ('args',))
            for setting in ["buffer_size", "buffer_timeout", "newline_re", "max_line_length"]:
                if setting not in msg["args"]:
                    raise KeyError('message did not contain obligatory settings for worker')

            self.factory.buildbot_bot.buffer_size = msg["args"]["buffer_size"]
            self.factory.buildbot_bot.buffer_timeout = msg["args"]["buffer_timeout"]
            self.factory.buildbot_bot.newline_re = msg["args"]["newline_re"]
            self.factory.buildbot_bot.max_line_length = msg["args"]["max_line_length"]
            result = None
        except Exception as e:
            is_exception = True
            result = str(e)

        self.send_response_msg(msg, result, is_exception)

    @defer.inlineCallbacks
    def call_start_command(self, msg):
        is_exception = False
        try:
            self.contains_msg_key(msg, ('command_id', 'command_name', 'args'))
            # send an instance, on which get_message_result will be called
            yield self.factory.buildbot_bot.start_command(
                self, msg['command_id'], msg['command_name'], msg['args']
            )
            result = None
        except Exception as e:
            is_exception = True
            result = str(e)

        self.send_response_msg(msg, result, is_exception)

    @defer.inlineCallbacks
    def call_shutdown(self, msg):
        is_exception = False
        try:
            yield self.factory.buildbot_bot.remote_shutdown()
            result = None
        except Exception as e:
            is_exception = True
            result = str(e)

        self.send_response_msg(msg, result, is_exception)

    @defer.inlineCallbacks
    def call_interrupt_command(self, msg):
        is_exception = False
        try:
            self.contains_msg_key(msg, ('command_id', 'why'))
            # send an instance, on which get_message_result will be called
            yield self.factory.buildbot_bot.interrupt_command(msg['command_id'], msg['why'])
            result = None
        except Exception as e:
            is_exception = True
            result = str(e)
        self.send_response_msg(msg, result, is_exception)

    def send_response_msg(self, msg, result, is_exception):
        dict_output = {'op': 'response', 'seq_number': msg['seq_number'], 'result': result}

        if is_exception:
            dict_output['is_exception'] = True
        self.maybe_log_worker_to_master_msg(dict_output)
        payload = msgpack.packb(dict_output)
        self.sendMessage(payload, isBinary=True)

    def onMessage(self, payload, isBinary):
        if not isBinary:
            log.msg('Message type form master unsupported')
            return

        msg = msgpack.unpackb(payload, raw=False)
        self.maybe_log_master_to_worker_msg(msg)

        if 'seq_number' not in msg or 'op' not in msg:
            log.msg(f'Invalid message from master: {msg}')
            return
        if msg['op'] == "print":
            self._deferwaiter.add(self.call_print(msg))
        elif msg['op'] == "keepalive":
            self._deferwaiter.add(self.call_keepalive(msg))
        elif msg['op'] == "set_worker_settings":
            self._deferwaiter.add(self.call_set_worker_settings(msg))
        elif msg['op'] == "get_worker_info":
            self._deferwaiter.add(self.call_get_worker_info(msg))
        elif msg['op'] == "start_command":
            self._deferwaiter.add(self.call_start_command(msg))
        elif msg['op'] == "shutdown":
            self._deferwaiter.add(self.call_shutdown(msg))
        elif msg['op'] == "interrupt_command":
            self._deferwaiter.add(self.call_interrupt_command(msg))
        elif msg['op'] == "response":
            seq_number = msg['seq_number']
            if "is_exception" in msg:
                self.seq_num_to_waiters_map[seq_number].errback(RemoteWorkerError(msg['result']))
            else:
                self.seq_num_to_waiters_map[seq_number].callback(msg['result'])
            # stop waiting for a response of this command
            del self.seq_num_to_waiters_map[seq_number]
        else:
            self.send_response_msg(
                msg, "Command {} does not exist.".format(msg['op']), is_exception=True
            )

    @defer.inlineCallbacks
    def get_message_result(self, msg):
        msg['seq_number'] = self.seq_number
        self.maybe_log_worker_to_master_msg(msg)
        msg = msgpack.packb(msg)
        d = defer.Deferred()
        self.seq_num_to_waiters_map[self.seq_number] = d
        self.seq_number = self.seq_number + 1
        self.sendMessage(msg, isBinary=True)
        res1 = yield d
        return res1

    def onClose(self, wasClean, code, reason):
        if self.debug:
            log.msg(f"WebSocket connection closed: {reason}")
        # stop waiting for the responses of all commands
        for seq_number in self.seq_num_to_waiters_map:
            self.seq_num_to_waiters_map[seq_number].errback(ConnectionLostError("Connection lost"))
        self.seq_num_to_waiters_map.clear()


class BuildbotWebSocketClientFactory(WebSocketClientFactory):
    def waitForCompleteShutdown(self):
        pass



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
#
# pylint: disable=C0111

import traceback
import warnings
from importlib.metadata import distributions
from importlib.metadata import entry_points

from zope.interface import Invalid
from zope.interface.verify import verifyClass

from buildbot.errors import PluginDBError
from buildbot.interfaces import IPlugin
from buildbot.util.importlib_compat import entry_points_get

# Base namespace for Buildbot specific plugins
_NAMESPACE_BASE = 'buildbot'


def find_distribution_info(entry_point_name, entry_point_group):
    for distribution in distributions():
        # each distribution can have many entry points
        try:
            for ep in entry_points_get(distribution.entry_points, entry_point_group):
                if ep.name == entry_point_name:
                    return (distribution.metadata['Name'], distribution.metadata['Version'])
        except KeyError as exc:
            raise PluginDBError("Plugin info was found, but it is invalid.") from exc
    raise PluginDBError("Plugin info not found.")


class _PluginEntry:
    def __init__(self, group, entry, loader):
        self._group = group
        self._entry = entry
        self._value = None
        self._loader = loader
        self._load_warnings = []
        self._info = None

    def load(self):
        if self._value is None:
            with warnings.catch_warnings(record=True) as all_warnings:
                warnings.simplefilter("always")
                self._value = self._loader(self._entry)
                self._load_warnings = list(all_warnings)

    @property
    def group(self):
        return self._group

    @property
    def name(self):
        return self._entry.name

    @property
    def info(self):
        if self._info is None:
            self._info = find_distribution_info(self._entry.name, self._group)
        return self._info

    def __eq__(self, other):
        return self.info == other.info

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def value(self):
        self.load()
        for w in self._load_warnings:
            warnings.warn_explicit(w.message, w.category, w.filename, w.lineno)
        return self._value


class _PluginEntryProxy(_PluginEntry):
    """Proxy for specific entry with custom group name.

    Used to provided access to the same entry from different namespaces.
    """

    def __init__(self, group, plugin_entry):
        assert isinstance(plugin_entry, _PluginEntry)
        self._plugin_entry = plugin_entry
        self._group = group

    def load(self):
        self._plugin_entry.load()

    @property
    def group(self):
        return self._group

    @property
    def name(self):
        return self._plugin_entry.name

    @property
    def info(self):
        return self._plugin_entry.info

    @property
    def value(self):
        return self._plugin_entry.value


class _NSNode:
    # pylint: disable=W0212

    def __init__(self):
        self._children = {}

    def load(self):
        for child in self._children.values():
            child.load()

    def add(self, name, entry):
        assert isinstance(name, str) and isinstance(entry, _PluginEntry)
        self._add(name, entry)

    def _add(self, name, entry):
        path = name.split('.', 1)
        key = path.pop(0)
        is_leaf = not path
        child = self._children.get(key)

        if is_leaf:
            if child is not None:
                assert isinstance(child, _PluginEntry)
                if child != entry:
                    raise PluginDBError(
                        f'Duplicate entry point for "{child.group}:{child.name}".\n'
                        f'  Previous definition {child.info}\n'
                        f'  This definition {entry.info}'
                    )
            else:
                self._children[key] = entry
        else:
            if child is None:
                child = _NSNode()
            assert isinstance(child, _NSNode)
            child._add(path[0], entry)
            self._children[key] = child

    def __getattr__(self, name):
        child = self._children.get(name)
        if child is None:
            raise PluginDBError(f'Unknown component name: {name}')

        if isinstance(child, _PluginEntry):
            return child.value
        return child

    def info(self, name):
        assert isinstance(name, str)

        return self._get(name).info

    def get(self, name):
        assert isinstance(name, str)

        return self._get(name).value

    def _get(self, name):
        path = name.split('.', 1)
        key = path.pop(0)
        is_leaf = not path
        child = self._children.get(key)

        if isinstance(child, _PluginEntry):
            if not is_leaf:
                raise PluginDBError(f'Excessive namespace specification: {path[0]}')
            return child
        elif child is None:
            raise PluginDBError(f'Unknown component name: {name}')
        else:
            return child._get(path[0])

    def _info_all(self):
        result = []
        for key, child in self._children.items():
            if isinstance(child, _PluginEntry):
                result.append((key, child.info))
            else:
                result.extend([
                    (f'{key}.{name}', value) for name, value in child.info_all().items()
                ])
        return result

    def info_all(self):
        return dict(self._info_all())


class _Plugins:
    """
    represent plugins within a namespace
    """

    def __init__(self, namespace, interface=None):
        if interface is not None:
            assert interface.isOrExtends(IPlugin)

        self._group = f'{_NAMESPACE_BASE}.{namespace}'

        self._interface = interface
        self._real_tree = None

    def _load_entry(self, entry):
        # pylint: disable=W0703
        try:
            result = entry.load()
        except Exception as e:
            # log full traceback of the bad entry to help support
            traceback.print_exc()
            raise PluginDBError(f'Unable to load {self._group}:{entry.name}: {e!s}') from e
        if self._interface:
            try:
                verifyClass(self._interface, result)
            except Invalid as e:
                raise PluginDBError(
                    f'Plugin {self._group}:{entry.name} does not implement '
                    f'{self._interface.__name__}: {e!s}'
                ) from e
        return result

    @property
    def _tree(self):
        if self._real_tree is None:
            self._real_tree = _NSNode()
            entries = entry_points_get(entry_points(), self._group)
            for entry in entries:
                self._real_tree.add(entry.name, _PluginEntry(self._group, entry, self._load_entry))
        return self._real_tree

    def load(self):
        self._tree.load()

    def info_all(self):
        return self._tree.info_all()

    @property
    def names(self):
        # Expensive operation
        return list(self.info_all())

    def info(self, name):
        """
        get information about a particular plugin if known in this namespace
        """
        return self._tree.info(name)

    def __contains__(self, name):
        """
        check if the given name is available as a plugin
        """
        try:
            return not isinstance(self._tree.get(name), _NSNode)
        except PluginDBError:
            return False

    def get(self, name):
        """
        get an instance of the plugin with the given name
        """
        return self._tree.get(name)

    def _get_entry(self, name):
        return self._tree._get(name)

    def __getattr__(self, name):
        try:
            return getattr(self._tree, name)
        except PluginDBError as e:
            raise AttributeError(str(e)) from e


class _PluginDB:
    """
    Plugin infrastructure support for Buildbot
    """

    def __init__(self):
        self._namespaces = {}

    def add_namespace(self, namespace, interface=None, load_now=False):
        """
        register given namespace in global database of plugins

        in case it's already registered, return the registration
        """
        tempo = self._namespaces.get(namespace)

        if tempo is None:
            tempo = _Plugins(namespace, interface)
            self._namespaces[namespace] = tempo

        if load_now:
            tempo.load()

        return tempo

    @property
    def namespaces(self):
        """
        get a list of registered namespaces
        """
        return list(self._namespaces)

    def info(self):
        """
        get information about all plugins in registered namespaces
        """
        result = {}
        for name, namespace in self._namespaces.items():
            result[name] = namespace.info_all()
        return result


_DB = _PluginDB()


def namespaces():
    """
    provide information about known namespaces
    """
    return _DB.namespaces


def info():
    """
    provide information about all known plugins

    format of the output:

    {<namespace>, {
        {<plugin-name>: (<package-name>, <package-version),
         ...},
        ...
    }
    """
    return _DB.info()


def get_plugins(namespace, interface=None, load_now=False):
    """
    helper to get a direct interface to _Plugins
    """
    return _DB.add_namespace(namespace, interface, load_now)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.util.render_description import render_description


class TestRaml(unittest.TestCase):
    def test_plain(self):
        self.assertIsNone(render_description("description", None))

    def test_unknown(self):
        with self.assertRaises(RuntimeError):
            render_description("description", "unknown")

    def test_markdown(self):
        self.assertEqual(
            render_description("# description\ntext", "markdown"),
            "<h1>description</h1>\n<p>text</p>",
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import asyncio
import multiprocessing
import os
import signal
import socket
import sys

from twisted.internet import defer

from buildbot.test.util.integration import RunMasterBase

from ..interop import test_commandmixin
from ..interop import test_compositestepmixin
from ..interop import test_integration_secrets
from ..interop import test_interruptcommand
from ..interop import test_setpropertyfromcommand
from ..interop import test_transfer
from ..interop import test_worker_reconnect

# This integration test puts HTTP proxy in between the master and worker.


def get_log_path():
    return f'test_worker_proxy_stdout_{os.getpid()}.txt'


def write_to_log(msg, with_traceback=False):
    with open(get_log_path(), 'a', encoding='utf-8') as outfile:
        outfile.write(msg)
        if with_traceback:
            import traceback

            traceback.print_exc(file=outfile)


async def handle_client(local_reader, local_writer):
    async def pipe(reader, writer):
        try:
            while not reader.at_eof():
                writer.write(await reader.read(2048))
        except ConnectionResetError:
            pass
        finally:
            writer.close()

    try:
        request = await local_reader.read(2048)
        lines = request.split(b"\r\n")
        if not lines[0].startswith(b"CONNECT "):
            write_to_log(f"bad request {request.decode()}\n")
            local_writer.write(b"HTTP/1.1 407 Only CONNECT allowed\r\n\r\n")
            return
        host, port = lines[0].split(b" ")[1].split(b":")
        try:
            remote_reader, remote_writer = await asyncio.open_connection(host.decode(), int(port))
        except socket.gaierror:
            write_to_log(f"failed to relay to {host} {port}\n")
            local_writer.write(b"HTTP/1.1 404 Not Found\r\n\r\n")
            return

        write_to_log(f"relaying to {host} {port}\n")
        local_writer.write(b"HTTP/1.1 200 Connection established\r\n\r\n")
        pipe1 = pipe(local_reader, remote_writer)
        pipe2 = pipe(remote_reader, local_writer)
        await asyncio.gather(pipe1, pipe2)

    finally:
        local_writer.close()


def run_proxy(queue):
    write_to_log("run_proxy\n")

    try:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # https://github.com/python/cpython/issues/83710
            if sys.version_info <= (3, 10, 8):
                # Workaround for bugs.python.org/issue39529.
                try:
                    loop = asyncio.get_event_loop_policy().get_event_loop()
                except RuntimeError:
                    # We can get RuntimeError due to current thread being not main thread
                    # on Python 3.8. It's not clear why that happens, so work around it.
                    loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(loop)
            else:
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)

        coro = asyncio.start_server(handle_client, host="127.0.0.1")
        server = loop.run_until_complete(coro)

        host, port = server.sockets[0].getsockname()

        queue.put(port)

        def signal_handler(sig, trace):
            raise KeyboardInterrupt

        signal.signal(signal.SIGTERM, signal_handler)

        write_to_log(f"Serving on {host}:{port}\n")
        try:
            write_to_log("Running forever\n")
            loop.run_forever()
        except KeyboardInterrupt:
            write_to_log("End\n")

        server.close()
        loop.run_until_complete(server.wait_closed())
        loop.close()

    except BaseException as e:
        write_to_log(f"Exception Raised: {e!s}\n", with_traceback=True)

    finally:
        queue.put(get_log_path())


class RunMasterBehindProxy(RunMasterBase):
    # we need slightly longer timeout for proxy related tests
    timeout = 30
    enable_debug = False

    def setUp(self):
        write_to_log("setUp\n")
        self.queue = multiprocessing.Queue()
        self.proxy_process = multiprocessing.Process(target=run_proxy, args=(self.queue,))
        self.proxy_process.start()
        self.target_port = self.queue.get()
        write_to_log(f"got target_port {self.target_port}\n")

        def cleanup():
            write_to_log("cleanup\n")
            self.proxy_process.terminate()
            self.proxy_process.join()
            if self.enable_debug:
                print("---- stdout ----")
                with open(get_log_path(), encoding='utf-8') as file:
                    print(file.read())
                print("---- ------ ----")
                with open(self.queue.get(), encoding='utf-8') as file:
                    print(file.read())
                print("---- ------ ----")
                os.unlink(get_log_path())

        self.addCleanup(cleanup)

    @defer.inlineCallbacks
    def setup_master(self, config_dict, startWorker=True):
        proxy_connection_string = f"tcp:127.0.0.1:{self.target_port}"
        yield super().setup_master(
            config_dict, startWorker, proxy_connection_string=proxy_connection_string
        )


# Use interoperability test cases to test the HTTP proxy tunneling.


class ProxyCommandMixinMasterPB(RunMasterBehindProxy, test_commandmixin.CommandMixinMasterPB):
    pass


class ProxyCompositeStepMixinMasterPb(
    RunMasterBehindProxy, test_compositestepmixin.CompositeStepMixinMasterPb
):
    pass


class ProxyInterruptCommandPb(RunMasterBehindProxy, test_interruptcommand.InterruptCommandPb):
    pass


class ProxySecretsConfigPB(RunMasterBehindProxy, test_integration_secrets.SecretsConfigPB):
    pass


class ProxySetPropertyFromCommandPB(
    RunMasterBehindProxy, test_setpropertyfromcommand.SetPropertyFromCommandPB
):
    pass


class ProxyTransferStepsMasterPb(RunMasterBehindProxy, test_transfer.TransferStepsMasterPb):
    pass


class ProxyWorkerReconnect(RunMasterBehindProxy, test_worker_reconnect.WorkerReconnectPb):
    pass



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

from typing import TYPE_CHECKING

from twisted.internet import defer

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable
    from typing import ClassVar
    from typing import Literal

    from typing_extensions import Self


class Expect:
    """
    An expected instantiation of RunProcess.  Usually used within a RunProcess
    expect invocation:

        rp.expect(
            Expect("echo", "bar", usePTY=False)
             + { 'stdout' : 'hello!!' }
             + { 'rc' : 13 }
             + 13 # for a callback with rc=13; or
             + Failure(..), # for a failure
            Expect(..) + .. ,
            ...
        )

    Note that the default values are accepted for all keyword arguments if they
    are not omitted.
    """

    def __init__(self, command: str | list[str], workdir: str, **kwargs: Any) -> None:
        self.kwargs: dict[str, Any] = {"command": command, "workdir": workdir}
        self.kwargs.update(kwargs)

        self.result: tuple[Literal['c'], int] | tuple[Literal['e'], BaseException] | None = None
        self.status_updates: list[list[tuple[str, Any]]] = []

    def __str__(self) -> str:
        other_kwargs = self.kwargs.copy()
        del other_kwargs['command']
        del other_kwargs['workdir']
        return "Command: {}\n  workdir: {}\n  kwargs: {}\n  result: {}\n".format(
            self.kwargs['command'], self.kwargs['workdir'], other_kwargs, self.result
        )

    def update(self, key: str, value: Any) -> Self:
        self.status_updates.append([(key, value)])
        return self

    def updates(self, updates: list[tuple[str, Any]]) -> Self:
        self.status_updates.append(updates)
        return self

    def exit(self, rc_code: int) -> Self:
        self.result = ('c', rc_code)
        return self

    def exception(self, error: BaseException) -> Self:
        self.result = ('e', error)
        return self


class FakeRunProcess:
    """
    A fake version of L{buildbot_worker.runprocess.RunProcess} which will
    simulate running external processes without actually running them (which is
    very fragile in tests!)

    This class is first programmed with the set of instances that are expected,
    and with their expected results.  It will raise an AssertionError if the
    expected behavior is not seen.

    Note that this handles sendStderr/sendStdout and keepStderr/keepStdout properly.
    """

    _expectations: ClassVar[list[Expect]]

    @classmethod
    def expect(cls, *expectations: Expect) -> None:
        """
        Set the expectations for this test run
        """
        cls._expectations = list(expectations)
        # list the first expectation last, so we can pop it
        cls._expectations.reverse()

    @classmethod
    def test_done(cls) -> None:
        """
        Indicate that this test is finished; if any expected instantiations
        have not taken place, this will raise the appropriate AssertionError.
        """
        if cls._expectations:
            raise AssertionError(f"{len(cls._expectations)} expected instances not created")
        del cls._expectations

    def __init__(
        self,
        command_id: int,
        command: list[str | bytes] | str | bytes,
        workdir: str,
        unicode_encoding: str,
        send_update: Callable[[list[tuple[str, Any]]], None],
        **kwargs: Any,
    ) -> None:
        kwargs['command'] = command
        kwargs['workdir'] = workdir

        # the default values for the constructor kwargs; if we got a default
        # value in **kwargs and didn't expect anything, well count that as OK
        default_values: dict[str, Any] = {
            "environ": None,
            "sendStdout": True,
            "sendStderr": True,
            "sendRC": True,
            "timeout": None,
            "maxTime": None,
            "max_lines": None,
            "sigtermTime": None,
            "initialStdin": None,
            "keepStdout": False,
            "keepStderr": False,
            "logEnviron": True,
            "logfiles": {},
            "usePTY": False,
        }

        if not self._expectations:
            raise AssertionError(f"unexpected instantiation: {kwargs}")
        exp = self._exp = self._expectations.pop()
        if exp.kwargs != kwargs:
            msg = []
            # pylint: disable=consider-iterating-dictionary
            for key in sorted(list(set(exp.kwargs.keys()) | set(kwargs.keys()))):
                if key not in exp.kwargs:
                    if key in default_values:
                        if default_values[key] == kwargs[key]:
                            continue  # default values are expected
                        msg.append(
                            f'{key}: expected default ({default_values[key]!r}),\n  got {kwargs[key]!r}'
                        )
                    else:
                        msg.append(f'{key}: unexpected arg, value = {kwargs[key]!r}')
                elif key not in kwargs:
                    msg.append(f'{key}: did not get expected arg')
                elif exp.kwargs[key] != kwargs[key]:
                    msg.append(f'{key}: expected {exp.kwargs[key]!r},\n  got {kwargs[key]!r}')
            if msg:
                msg.insert(
                    0,
                    'did not get expected __init__ arguments for\n {}'.format(
                        " ".join(map(repr, kwargs.get('command', ['unknown command'])))
                    ),
                )
                self._expectations[:] = []  # don't expect any more instances, since we're failing
                raise AssertionError("\n".join(msg))

        self.send_update = send_update
        self.stdout = ''
        self.stderr = ''

        self.run_deferred: defer.Deferred[int] | None = None

    def start(self) -> defer.Deferred[int]:
        # figure out the stdio-related parameters
        keepStdout = self._exp.kwargs.get('keepStdout', False)
        keepStderr = self._exp.kwargs.get('keepStderr', False)
        sendStdout = self._exp.kwargs.get('sendStdout', True)
        sendStderr = self._exp.kwargs.get('sendStderr', True)
        if keepStdout:
            self.stdout = ''
        if keepStderr:
            self.stderr = ''
        finish_immediately = True

        for update in self._exp.status_updates:
            data = []
            for key, value in update:
                if key == 'stdout':
                    if keepStdout:
                        self.stdout += value
                    if not sendStdout:
                        continue  # don't send this update
                if key == 'stderr':
                    if keepStderr:
                        self.stderr += value
                    if not sendStderr:
                        continue
                if key == 'wait':
                    finish_immediately = False
                    continue
                data.append((key, value))
            self.send_update(data)

        d = self.run_deferred = defer.Deferred()

        if finish_immediately:
            self._finished()

        return d

    def _finished(self) -> None:
        assert self._exp.result and self._exp.result[0] is not None
        assert self.run_deferred is not None
        if self._exp.result[0] == 'e':
            self.run_deferred.errback(self._exp.result[1])
        else:
            self.run_deferred.callback(self._exp.result[1])

    def kill(self, reason: str) -> None:
        self.send_update([('header', 'killing')])
        self.send_update([('rc', -1)])
        assert self.run_deferred is not None
        self.run_deferred.callback(-1)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.db import masters
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util import epoch2datetime

SOMETIME = 1348971992
SOMETIME_DT = epoch2datetime(SOMETIME)
OTHERTIME = 1008971992
OTHERTIME_DT = epoch2datetime(OTHERTIME)


class Tests(TestReactorMixin, unittest.TestCase):
    # common sample data

    master_row = [
        fakedb.Master(id=7, active=1, last_active=SOMETIME),
    ]

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.reactor.advance(SOMETIME)
        self.master = yield fakemaster.make_master(self, wantDb=True)
        self.db = self.master.db

    @defer.inlineCallbacks
    def test_findMasterId_new(self):
        id = yield self.db.masters.findMasterId('master-7')
        masterdict = yield self.db.masters.getMaster(id)
        self.assertEqual(
            masterdict,
            masters.MasterModel(id=id, name='master-7', active=False, last_active=SOMETIME_DT),
        )

    @defer.inlineCallbacks
    def test_findMasterId_new_name_differs_only_by_case(self):
        yield self.db.insert_test_data([
            fakedb.Master(id=7, name='some:master'),
        ])
        id = yield self.db.masters.findMasterId('some:Master')
        masterdict = yield self.db.masters.getMaster(id)
        self.assertEqual(
            masterdict,
            masters.MasterModel(id=id, name='some:Master', active=False, last_active=SOMETIME_DT),
        )

    @defer.inlineCallbacks
    def test_findMasterId_exists(self):
        yield self.db.insert_test_data([
            fakedb.Master(id=7, name='some:master'),
        ])
        id = yield self.db.masters.findMasterId('some:master')
        self.assertEqual(id, 7)

    @defer.inlineCallbacks
    def test_setMasterState_when_missing(self):
        activated = yield self.db.masters.setMasterState(masterid=7, active=True)
        self.assertFalse(activated)

    @defer.inlineCallbacks
    def test_setMasterState_true_when_active(self):
        yield self.db.insert_test_data([
            fakedb.Master(id=7, active=1, last_active=OTHERTIME),
        ])
        activated = yield self.db.masters.setMasterState(masterid=7, active=True)
        self.assertFalse(activated)  # it was already active
        masterdict = yield self.db.masters.getMaster(7)
        self.assertEqual(
            masterdict,
            masters.MasterModel(id=7, name='master-7', active=True, last_active=SOMETIME_DT),
        )  # timestamp updated

    @defer.inlineCallbacks
    def test_setMasterState_true_when_inactive(self):
        yield self.db.insert_test_data([
            fakedb.Master(id=7, active=0, last_active=OTHERTIME),
        ])
        activated = yield self.db.masters.setMasterState(masterid=7, active=True)
        self.assertTrue(activated)
        masterdict = yield self.db.masters.getMaster(7)
        self.assertEqual(
            masterdict,
            masters.MasterModel(id=7, name='master-7', active=True, last_active=SOMETIME_DT),
        )

    @defer.inlineCallbacks
    def test_setMasterState_false_when_active(self):
        yield self.db.insert_test_data([
            fakedb.Master(id=7, active=1, last_active=OTHERTIME),
        ])
        deactivated = yield self.db.masters.setMasterState(masterid=7, active=False)
        self.assertTrue(deactivated)
        masterdict = yield self.db.masters.getMaster(7)
        self.assertEqual(
            masterdict,
            masters.MasterModel(id=7, name='master-7', active=False, last_active=OTHERTIME_DT),
        )

    @defer.inlineCallbacks
    def test_setMasterState_false_when_inactive(self):
        yield self.db.insert_test_data([
            fakedb.Master(id=7, active=0, last_active=OTHERTIME),
        ])
        deactivated = yield self.db.masters.setMasterState(masterid=7, active=False)
        self.assertFalse(deactivated)
        masterdict = yield self.db.masters.getMaster(7)
        self.assertEqual(
            masterdict,
            masters.MasterModel(id=7, name='master-7', active=False, last_active=OTHERTIME_DT),
        )

    @defer.inlineCallbacks
    def test_getMaster(self):
        yield self.db.insert_test_data([
            fakedb.Master(id=7, active=0, last_active=SOMETIME),
        ])
        masterdict = yield self.db.masters.getMaster(7)
        self.assertIsInstance(masterdict, masters.MasterModel)
        self.assertEqual(
            masterdict,
            masters.MasterModel(id=7, name='master-7', active=False, last_active=SOMETIME_DT),
        )

    @defer.inlineCallbacks
    def test_getMaster_missing(self):
        masterdict = yield self.db.masters.getMaster(7)
        self.assertEqual(masterdict, None)

    @defer.inlineCallbacks
    def test_getMasters(self):
        yield self.db.insert_test_data([
            fakedb.Master(id=7, active=0, last_active=SOMETIME),
            fakedb.Master(id=8, active=1, last_active=OTHERTIME),
        ])
        masterlist = yield self.db.masters.getMasters()
        for masterdict in masterlist:
            self.assertIsInstance(masterdict, masters.MasterModel)

        def masterKey(master):
            return master.id

        expected = sorted(
            [
                masters.MasterModel(id=7, name='master-7', active=0, last_active=SOMETIME_DT),
                masters.MasterModel(id=8, name='master-8', active=1, last_active=OTHERTIME_DT),
            ],
            key=masterKey,
        )
        self.assertEqual(sorted(masterlist, key=masterKey), expected)

    @defer.inlineCallbacks
    def test_setMasterState_false_deletes_links(self):
        yield self.db.insert_test_data([
            fakedb.Master(id=7, name='some:master', active=1, last_active=OTHERTIME),
            fakedb.Scheduler(id=21),
            fakedb.SchedulerMaster(schedulerid=21, masterid=7),
        ])
        deactivated = yield self.db.masters.setMasterState(masterid=7, active=False)
        self.assertTrue(deactivated)

        # check that the scheduler_masters row was deleted
        def thd(conn):
            tbl = self.db.model.scheduler_masters
            self.assertEqual(conn.execute(tbl.select()).fetchall(), [])

        yield self.db.pool.do(thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.process.properties import WithProperties
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.steps import cppcheck
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import TestBuildStepMixin


class Cppcheck(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_success(self):
        self.setup_step(cppcheck.Cppcheck(enable=['all'], inconclusive=True))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir', command=['cppcheck', '.', '--enable=all', '--inconclusive']
            )
            .stdout('Checking file1.c...')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="cppcheck")
        return self.run_step()

    def test_command_failure(self):
        self.setup_step(cppcheck.Cppcheck(enable=['all'], inconclusive=True))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir', command=['cppcheck', '.', '--enable=all', '--inconclusive']
            )
            .stdout('Checking file1.c...')
            .exit(1)
        )
        self.expect_outcome(result=FAILURE, state_string="cppcheck (failure)")
        return self.run_step()

    def test_warnings(self):
        self.setup_step(cppcheck.Cppcheck(source=['file1.c'], enable=['warning', 'performance']))
        self.expect_commands(
            ExpectShell(
                workdir='wkdir', command=['cppcheck', 'file1.c', '--enable=warning,performance']
            )
            .stdout(
                'Checking file1.c...\n'
                '[file1.c:3]: (warning) Logical disjunction always '
                'evaluates to true: t >= 0 || t < 65.\n'
                '(information) Cppcheck cannot find all the include files '
                '(use --check-config for details)'
            )
            .exit(0)
        )
        self.expect_outcome(
            result=WARNINGS, state_string="cppcheck warning=1 information=1 (warnings)"
        )
        return self.run_step()

    def test_errors(self):
        self.setup_step(cppcheck.Cppcheck(extra_args=['--my-param=5']))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['cppcheck', '.', '--my-param=5'])
            .stdout(
                'Checking file1.c...\n'
                '[file1.c:3]: (error) Possible null pointer dereference: filter\n'
                '[file1.c:4]: (error) Memory leak: columns\n'
                "[file1.c:7]: (style) The scope of the variable 'pid' can be reduced"
            )
            .exit(0)
        )
        self.expect_outcome(result=FAILURE, state_string="cppcheck error=2 style=1 (failure)")
        return self.run_step()

    def test_renderables(self):
        P = WithProperties
        self.setup_step(
            cppcheck.Cppcheck(
                binary=P('a'), source=[P('.'), P('f.c')], extra_args=[P('--p'), P('--p')]
            )
        )
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['a', '.', 'f.c', '--p', '--p'])
            .stdout('Checking file1.c...')
            .exit(0)
        )
        self.expect_outcome(result=SUCCESS, state_string="cppcheck")
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.python import log
from twisted.trial import unittest

from buildbot.process import results


class TestResults(unittest.TestCase):
    def test_Results(self):
        for r in results.Results:
            i = getattr(results, r.upper())
            self.assertEqual(results.Results[i], r)

    def test_worst_status(self):
        self.assertEqual(results.WARNINGS, results.worst_status(results.SUCCESS, results.WARNINGS))
        self.assertEqual(
            results.CANCELLED, results.worst_status(results.SKIPPED, results.CANCELLED)
        )

    def test_sort_worst_status(self):
        res = list(range(len(results.Results)))
        res.sort(key=lambda a: a if a != results.SKIPPED else -1)
        self.assertEqual(
            res,
            [
                results.SKIPPED,
                results.SUCCESS,
                results.WARNINGS,
                results.FAILURE,
                results.EXCEPTION,
                results.RETRY,
                results.CANCELLED,
            ],
        )

    def do_test_carc(
        self,
        result,
        previousResult,
        newResult,
        terminate,
        haltOnFailure=None,
        flunkOnWarnings=None,
        flunkOnFailure=None,
        warnOnWarnings=None,
        warnOnFailure=None,
    ):
        if haltOnFailure is None:
            haltOnFailure = [True, False]
        if flunkOnWarnings is None:
            flunkOnWarnings = [True, False]
        if flunkOnFailure is None:
            flunkOnFailure = [True, False]
        if warnOnWarnings is None:
            warnOnWarnings = [True, False]
        if warnOnFailure is None:
            warnOnFailure = [True, False]
        for hof in haltOnFailure:
            for fow in flunkOnWarnings:
                for fof in flunkOnFailure:
                    for wow in warnOnWarnings:
                        for wof in warnOnFailure:
                            self.haltOnFailure = hof
                            self.flunkOnWarnings = fow
                            self.flunkOnFailure = fof
                            self.warnOnWarnings = wow
                            self.warnOnFailure = wof
                            nr, term = results.computeResultAndTermination(
                                self, result, previousResult
                            )
                            log.msg(
                                f"res={results.Results[result]!r} "
                                f"prevRes={results.Results[previousResult]!r} "
                                f"hof={hof!r} fow={fow!r} fof={fof!r} "
                                f"wow={wow!r} wof={wof!r} => "
                                f"{results.Results[nr]!r} {term!r}"
                            )
                            self.assertEqual(
                                (nr, term), (newResult, terminate), "see test.log for details"
                            )

    def test_carc_success_after_success(self):
        self.do_test_carc(results.SUCCESS, results.SUCCESS, results.SUCCESS, False)

    def test_carc_success_after_warnings(self):
        self.do_test_carc(results.SUCCESS, results.WARNINGS, results.WARNINGS, False)

    def test_carc_success_after_failure(self):
        self.do_test_carc(results.SUCCESS, results.FAILURE, results.FAILURE, False)

    def test_carc_warnings_after_success(self):
        self.do_test_carc(
            results.WARNINGS,
            results.SUCCESS,
            results.WARNINGS,
            False,
            flunkOnWarnings=[False],
            warnOnWarnings=[True],
        )
        self.do_test_carc(
            results.WARNINGS,
            results.SUCCESS,
            results.SUCCESS,
            False,
            flunkOnWarnings=[False],
            warnOnWarnings=[False],
        )
        self.do_test_carc(
            results.WARNINGS,
            results.SUCCESS,
            results.FAILURE,
            False,
            flunkOnWarnings=[True],
            warnOnWarnings=[True],
        )
        self.do_test_carc(
            results.WARNINGS,
            results.SUCCESS,
            results.FAILURE,
            False,
            flunkOnWarnings=[True],
            warnOnWarnings=[False],
        )

    def test_carc_warnings_after_warnings(self):
        self.do_test_carc(
            results.WARNINGS, results.WARNINGS, results.WARNINGS, False, flunkOnWarnings=[False]
        )
        self.do_test_carc(
            results.WARNINGS, results.WARNINGS, results.FAILURE, False, flunkOnWarnings=[True]
        )

    def test_carc_warnings_after_failure(self):
        self.do_test_carc(
            results.WARNINGS, results.FAILURE, results.FAILURE, False, flunkOnWarnings=[False]
        )
        self.do_test_carc(
            results.WARNINGS, results.FAILURE, results.FAILURE, False, flunkOnWarnings=[True]
        )

    def test_carc_failure_after_success(self):
        for hof in False, True:
            self.do_test_carc(
                results.FAILURE,
                results.SUCCESS,
                results.FAILURE,
                hof,
                haltOnFailure=[hof],
                flunkOnFailure=[True],
                warnOnFailure=[False],
            )
            self.do_test_carc(
                results.FAILURE,
                results.SUCCESS,
                results.FAILURE,
                hof,
                haltOnFailure=[hof],
                flunkOnFailure=[True],
                warnOnFailure=[True],
            )
            self.do_test_carc(
                results.FAILURE,
                results.SUCCESS,
                results.SUCCESS,
                hof,
                haltOnFailure=[hof],
                flunkOnFailure=[False],
                warnOnFailure=[False],
            )
            self.do_test_carc(
                results.FAILURE,
                results.SUCCESS,
                results.WARNINGS,
                hof,
                haltOnFailure=[hof],
                flunkOnFailure=[False],
                warnOnFailure=[True],
            )

    def test_carc_failure_after_warnings(self):
        for hof in False, True:
            self.do_test_carc(
                results.FAILURE,
                results.WARNINGS,
                results.FAILURE,
                hof,
                haltOnFailure=[hof],
                flunkOnFailure=[True],
            )
            self.do_test_carc(
                results.FAILURE,
                results.WARNINGS,
                results.WARNINGS,
                hof,
                haltOnFailure=[hof],
                flunkOnFailure=[False],
            )

    def test_carc_failure_after_failure(self):
        for hof in False, True:
            self.do_test_carc(
                results.FAILURE, results.FAILURE, results.FAILURE, hof, haltOnFailure=[hof]
            )

    def test_carc_exception(self):
        for prev in results.FAILURE, results.WARNINGS, results.SUCCESS:
            self.do_test_carc(results.EXCEPTION, prev, results.EXCEPTION, True)

    def test_carc_retry(self):
        for prev in results.FAILURE, results.WARNINGS, results.SUCCESS:
            self.do_test_carc(results.RETRY, prev, results.RETRY, True)

    def test_carc_cancelled(self):
        for prev in results.FAILURE, results.WARNINGS, results.SUCCESS:
            self.do_test_carc(results.CANCELLED, prev, results.CANCELLED, True)

    def test_carc_skipped(self):
        for prev in results.FAILURE, results.WARNINGS, results.SUCCESS:
            self.do_test_carc(results.SKIPPED, prev, prev, False)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import sqlalchemy as sa
from twisted.trial import unittest

from buildbot.test.util import migration
from buildbot.util import sautils


class Migration(migration.MigrateTestMixin, unittest.TestCase):
    def setUp(self):
        return self.setUpMigrateTest()

    def create_tables_thd(self, conn):
        metadata = sa.MetaData()
        metadata.bind = conn

        # buildid foreign key is removed for the purposes of the test
        steps = sautils.Table(
            'steps',
            metadata,
            sa.Column('id', sa.Integer, primary_key=True),
            sa.Column('number', sa.Integer, nullable=False),
            sa.Column('name', sa.String(50), nullable=False),
            sa.Column('buildid', sa.Integer, nullable=True),
            sa.Column('started_at', sa.Integer),
            sa.Column('complete_at', sa.Integer),
            sa.Column('state_string', sa.Text, nullable=False),
            sa.Column('results', sa.Integer),
            sa.Column('urls_json', sa.Text, nullable=False),
            sa.Column('hidden', sa.SmallInteger, nullable=False, server_default='0'),
        )
        steps.create(bind=conn)

        conn.execute(
            steps.insert(),
            [
                {
                    "id": 4,
                    "number": 123,
                    "name": "step",
                    "buildid": 12,
                    "started_at": 1690848000,
                    "complete_at": 1690848030,
                    "state_string": "state",
                    "results": 0,
                    "urls_json": "",
                    "hidden": 0,
                }
            ],
        )
        conn.commit()

    def test_update(self):
        def setup_thd(conn):
            self.create_tables_thd(conn)

        def verify_thd(conn):
            metadata = sa.MetaData()
            metadata.bind = conn

            steps = sautils.Table('steps', metadata, autoload_with=conn)
            self.assertIsInstance(steps.c.locks_acquired_at.type, sa.Integer)

            q = sa.select(
                steps.c.name,
                steps.c.locks_acquired_at,
            )

            num_rows = 0
            for row in conn.execute(q):
                self.assertEqual(row.name, "step")
                self.assertEqual(row.locks_acquired_at, 1690848000)
                num_rows += 1
            self.assertEqual(num_rows, 1)

        return self.do_test_migration('062', '063', setup_thd, verify_thd)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os
import shutil
from typing import TYPE_CHECKING

from twisted.internet import defer
from twisted.trial import unittest

if TYPE_CHECKING:
    from twisted.trial import unittest

    _DirsMixinBase = unittest.TestCase
else:
    _DirsMixinBase = object


class DirsMixin(_DirsMixinBase):
    _dirs = None

    def setUpDirs(self, *dirs) -> defer.Deferred[None]:
        """Make sure C{dirs} exist and are empty, and set them up to be deleted
        in tearDown."""
        self._dirs = map(os.path.abspath, dirs)
        for dir in self._dirs:
            if os.path.exists(dir):
                shutil.rmtree(dir)
            os.makedirs(dir)

        def cleanup():
            for dir in self._dirs:
                if os.path.exists(dir):
                    shutil.rmtree(dir)

        self.addCleanup(cleanup)

        # return a deferred to make chaining easier
        return defer.succeed(None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.internet import defer

from buildbot.process.properties import Interpolate
from buildbot.test.fake.secrets import FakeSecretStorage
from buildbot.test.util.integration import RunMasterBase


class SecretsConfig(RunMasterBase):
    @defer.inlineCallbacks
    def setup_config(self, use_with=False):
        c = {}
        from buildbot.config import BuilderConfig
        from buildbot.plugins import schedulers
        from buildbot.plugins import steps
        from buildbot.process.factory import BuildFactory

        c['schedulers'] = [schedulers.ForceScheduler(name="force", builderNames=["testy"])]

        c['secretsProviders'] = [FakeSecretStorage(secretdict={"foo": "bar", "something": "more"})]
        f = BuildFactory()
        if use_with:
            secrets_list = [("pathA", Interpolate('%(secret:something)s'))]
            with f.withSecrets(secrets_list):
                f.addStep(steps.ShellCommand(command=Interpolate('echo %(secret:foo)s')))
        else:
            f.addSteps(
                [steps.ShellCommand(command=Interpolate('echo %(secret:foo)s'))],
                withSecrets=[("pathA", Interpolate('%(secret:something)s'))],
            )
        c['builders'] = [BuilderConfig(name="testy", workernames=["local1"], factory=f)]
        yield self.setup_master(c)

    @defer.inlineCallbacks
    def test_secret(self):
        yield self.setup_config()
        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        res = yield self.checkBuildStepLogExist(build, "<foo>")
        self.assertTrue(res)

    @defer.inlineCallbacks
    def test_withsecrets(self):
        yield self.setup_config(use_with=True)
        build = yield self.doForceBuild(wantSteps=True, wantLogs=True)
        self.assertEqual(build['buildid'], 1)
        res = yield self.checkBuildStepLogExist(build, "<foo>")
        self.assertTrue(res)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import datetime
import os
import shutil
import subprocess
from pathlib import Path


class TestGitRepository:
    def __init__(self, repository_path: os.PathLike, git_bin: os.PathLike | None | str = None):
        if git_bin is None:
            git_bin = shutil.which('git')
            if git_bin is None:
                raise FileNotFoundError('Failed to find git')

        self.git_bin = git_bin

        self.repository_path = Path(repository_path)
        self.repository_path.mkdir(parents=True, exist_ok=True)

        self.curr_date = datetime.datetime(2024, 6, 8, 14, 0, 0, tzinfo=datetime.timezone.utc)
        self.curr_author_name = 'test user'
        self.curr_author_email = 'user@example.com'

        self.exec_git(['init', '--quiet', '--initial-branch=main'])

    def advance_time(self, timedelta):
        self.curr_date += timedelta

    def create_file_text(self, relative_path: str, contents: str):
        path = self.repository_path / relative_path
        path.write_text(contents)
        os.utime(path, (self.curr_date.timestamp(), self.curr_date.timestamp()))

    def amend_file_text(self, relative_path: str, contents: str):
        path = self.repository_path / relative_path
        with path.open('a') as fp:
            fp.write(contents)
        os.utime(path, (self.curr_date.timestamp(), self.curr_date.timestamp()))

    def exec_git(self, args: list[str], env: dict[str, str] | None = None):
        final_env = self.git_author_env(
            author_name=self.curr_author_name, author_mail=self.curr_author_email
        )
        final_env.update(self.git_date_env(self.curr_date))
        if env is not None:
            final_env.update(env)

        subprocess.check_call(
            [str(self.git_bin), *args],
            cwd=self.repository_path,
            env=final_env,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )

    def commit(
        self,
        message: str,
        files: list[os.PathLike] | None = None,
        env: dict[str, str] | None = None,
    ) -> str:
        args = ['commit', '--quiet', f'--message={message}']
        if files is not None:
            args.extend(str(f) for f in files)

        self.exec_git(args, env=env)

        return subprocess.check_output(
            [str(self.git_bin), 'rev-parse', 'HEAD'],
            cwd=self.repository_path,
            text=True,
        ).strip()

    @staticmethod
    def git_author_env(author_name: str, author_mail: str):
        return {
            "GIT_AUTHOR_NAME": author_name,
            "GIT_AUTHOR_EMAIL": author_mail,
            "GIT_COMMITTER_NAME": author_name,
            "GIT_COMMITTER_EMAIL": author_mail,
        }

    @staticmethod
    def git_date_env(date: datetime.datetime):
        def _format_date(_d: datetime.datetime) -> str:
            # just in case, make sure we use UTC
            return _d.astimezone(tz=datetime.timezone.utc).strftime("%Y-%m-%d %H:%M:%S +0000")

        return {
            "GIT_AUTHOR_DATE": _format_date(date),
            "GIT_COMMITTER_DATE": _format_date(date),
        }



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.changes.changes import Change
from buildbot.process.properties import Properties
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.steps.source import repo
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectMkdir
from buildbot.test.steps import ExpectRmdir
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import ExpectStat
from buildbot.test.util import sourcesteps


class RepoURL(unittest.TestCase):
    # testcases taken from old_source/Repo test

    def oneTest(self, props, expected):
        p = Properties()
        p.update(props, "test")
        r = repo.RepoDownloadsFromProperties(list(props))
        self.assertEqual(sorted(r.getRenderingFor(p)), sorted(expected))

    def test_parse1(self):
        self.oneTest({'a': "repo download test/bla 564/12"}, ["test/bla 564/12"])

    def test_parse2(self):
        self.oneTest(
            {'a': "repo download test/bla 564/12 repo download test/bla 564/2"},
            ["test/bla 564/12", "test/bla 564/2"],
        )
        self.oneTest(
            {'a': "repo download test/bla 564/12", 'b': "repo download test/bla 564/2"},
            ["test/bla 564/12", "test/bla 564/2"],
        )

    def test_parse3(self):
        self.oneTest(
            {'a': "repo download test/bla 564/12 repo download test/bla 564/2 test/foo 5/1"},
            ["test/bla 564/12", "test/bla 564/2", "test/foo 5/1"],
        )
        self.oneTest({'a': "repo download test/bla 564/12"}, ["test/bla 564/12"])


class TestRepo(sourcesteps.SourceStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        self.shouldRetry = False
        self.logEnviron = True
        return self.setup_test_build_step()

    def shouldLogEnviron(self):
        r = self.logEnviron
        self.logEnviron = False
        return r

    def ExpectShell(self, **kw):
        if 'workdir' not in kw:
            kw['workdir'] = 'wkdir'
        if 'log_environ' not in kw:
            kw['log_environ'] = self.shouldLogEnviron()
        return ExpectShell(**kw)

    def mySetupStep(self, **kwargs):
        if "repoDownloads" not in kwargs:
            kwargs.update({
                "repoDownloads": repo.RepoDownloadsFromProperties([
                    "repo_download",
                    "repo_download2",
                ])
            })
        self.setup_step(
            repo.Repo(
                manifestURL='git://myrepo.com/manifest.git',
                manifestBranch="mb",
                manifestFile="mf",
                **kwargs,
            )
        )
        self.build.allChanges = lambda x=None: []

    def myRunStep(self, result=SUCCESS, state_string=None):
        self.expect_outcome(result=result, state_string=state_string)
        return self.run_step()

    def expectClobber(self):
        # stat return 1 so we clobber
        self.expect_commands(
            ExpectStat(file='wkdir/.repo', log_environ=self.logEnviron).exit(1),
            ExpectRmdir(dir='wkdir', log_environ=self.logEnviron).exit(0),
            ExpectMkdir(dir='wkdir', log_environ=self.logEnviron).exit(0),
        )

    def expectnoClobber(self):
        # stat return 0, so nothing
        self.expect_commands(ExpectStat(file='wkdir/.repo', log_environ=self.logEnviron).exit(0))

    def expectRepoSync(
        self,
        which_fail=-1,
        breakatfail=False,
        depth=0,
        initoptions=None,
        syncoptions=None,
        override_commands=None,
    ):
        if initoptions is None:
            initoptions = []
        if syncoptions is None:
            syncoptions = ["-c"]
        if override_commands is None:
            override_commands = []
        commands = [
            self.ExpectShell(command=["bash", "-c", self.get_nth_step(0)._getCleanupCommand()]),
            self.ExpectShell(
                command=[
                    "repo",
                    "init",
                    "-u",
                    "git://myrepo.com/manifest.git",
                    "-b",
                    "mb",
                    "-m",
                    "mf",
                    "--depth",
                    str(depth),
                    *initoptions,
                ]
            ),
            *override_commands,
            self.ExpectShell(command=["repo", "sync", "--force-sync", *syncoptions]),
            self.ExpectShell(command=["repo", "manifest", "-r", "-o", "manifest-original.xml"]),
        ]
        for i, command in enumerate(commands):
            self.expect_commands(command.exit((which_fail == i and 1) or 0))
            if which_fail == i and breakatfail:
                break

    def test_basic(self):
        """basic first time repo sync"""
        self.mySetupStep(repoDownloads=None)
        self.expectClobber()
        self.expectRepoSync()
        return self.myRunStep()

    def test_basic_depth(self):
        """basic first time repo sync"""
        self.mySetupStep(repoDownloads=None, depth=2)
        self.expectClobber()
        self.expectRepoSync(depth=2)
        return self.myRunStep()

    def test_basic_submodule(self):
        """basic first time repo sync with submodule"""
        self.mySetupStep(repoDownloads=None, submodules=True)
        self.expectClobber()
        self.expectRepoSync(initoptions=["--submodules"])
        return self.myRunStep()

    def test_update(self):
        """basic second time repo sync"""
        self.mySetupStep()
        self.expectnoClobber()
        self.expectRepoSync()
        return self.myRunStep()

    def test_jobs(self):
        """basic first time repo sync with jobs"""
        self.mySetupStep(jobs=2)
        self.expectClobber()
        self.expectRepoSync(syncoptions=["-j2", "-c"])
        return self.myRunStep()

    def test_sync_all_branches(self):
        """basic first time repo sync with all branches"""
        self.mySetupStep(syncAllBranches=True)
        self.expectClobber()
        self.expectRepoSync(syncoptions=[])
        return self.myRunStep()

    def test_manifest_override(self):
        """repo sync with manifest_override_url property set
        download via wget
        """
        self.mySetupStep(manifestOverrideUrl="http://u.rl/test.manifest", syncAllBranches=True)
        self.expectClobber()
        override_commands = [
            ExpectStat(file='wkdir/http://u.rl/test.manifest', log_environ=False),
            self.ExpectShell(
                log_environ=False,
                command=['wget', 'http://u.rl/test.manifest', '-O', 'manifest_override.xml'],
            ),
            self.ExpectShell(
                log_environ=False,
                workdir='wkdir/.repo',
                command=['ln', '-sf', '../manifest_override.xml', 'manifest.xml'],
            ),
        ]
        self.expectRepoSync(which_fail=2, syncoptions=[], override_commands=override_commands)
        return self.myRunStep()

    def test_manifest_override_local(self):
        """repo sync with manifest_override_url property set
        copied from local FS
        """
        self.mySetupStep(manifestOverrideUrl="test.manifest", syncAllBranches=True)
        self.expectClobber()
        override_commands = [
            ExpectStat(file='wkdir/test.manifest', log_environ=False),
            self.ExpectShell(
                log_environ=False, command=['cp', '-f', 'test.manifest', 'manifest_override.xml']
            ),
            self.ExpectShell(
                log_environ=False,
                workdir='wkdir/.repo',
                command=['ln', '-sf', '../manifest_override.xml', 'manifest.xml'],
            ),
        ]
        self.expectRepoSync(syncoptions=[], override_commands=override_commands)
        return self.myRunStep()

    def test_tarball(self):
        """repo sync using the tarball cache"""
        self.mySetupStep(tarball="/tarball.tar")
        self.expectClobber()
        self.expect_commands(self.ExpectShell(command=['tar', '-xvf', '/tarball.tar']).exit(0))
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['stat', '-c%Y', '/tarball.tar']).stdout(str(10000)).exit(0)
        )
        self.expect_commands(
            self.ExpectShell(command=['stat', '-c%Y', '.'])
            .stdout(str(10000 + 7 * 24 * 3600))
            .exit(0)
        )
        return self.myRunStep()

    def test_create_tarball(self):
        """repo sync create the tarball if its not here"""
        self.mySetupStep(tarball="/tarball.tgz")
        self.expectClobber()
        self.expect_commands(
            self.ExpectShell(command=['tar', '-z', '-xvf', '/tarball.tgz']).exit(1),
            self.ExpectShell(command=['rm', '-f', '/tarball.tgz']).exit(1),
            ExpectRmdir(dir='wkdir/.repo', log_environ=False).exit(1),
        )
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['stat', '-c%Y', '/tarball.tgz'])
            .stderr("file not found!")
            .exit(1),
            self.ExpectShell(command=['tar', '-z', '-cvf', '/tarball.tgz', '.repo']).exit(0),
        )
        return self.myRunStep()

    def do_test_update_tarball(self, suffix, option):
        """repo sync update the tarball cache at the end (tarball older than a week)"""
        self.mySetupStep(tarball="/tarball." + suffix)
        self.expectClobber()
        self.expect_commands(
            self.ExpectShell(command=["tar", *option, "-xvf", "/tarball." + suffix]).exit(0)
        )
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['stat', '-c%Y', '/tarball.' + suffix])
            .stdout(str(10000))
            .exit(0),
            self.ExpectShell(command=['stat', '-c%Y', '.'])
            .stdout(str(10001 + 7 * 24 * 3600))
            .exit(0),
            self.ExpectShell(command=["tar", *option, "-cvf", "/tarball." + suffix, ".repo"]).exit(
                0
            ),
        )
        return self.myRunStep()

    def test_update_tarball(self):
        self.do_test_update_tarball("tar", [])

    def test_update_tarball_gz(self):
        """tarball compression variants"""
        self.do_test_update_tarball("tar.gz", ["-z"])

    def test_update_tarball_tgz(self):
        self.do_test_update_tarball("tgz", ["-z"])

    def test_update_tarball_pigz(self):
        self.do_test_update_tarball("pigz", ["-I", "pigz"])

    def test_update_tarball_bzip(self):
        self.do_test_update_tarball("tar.bz2", ["-j"])

    def test_update_tarball_lzma(self):
        self.do_test_update_tarball("tar.lzma", ["--lzma"])

    def test_update_tarball_lzop(self):
        self.do_test_update_tarball("tar.lzop", ["--lzop"])

    def test_update_tarball_fail1(self, suffix="tar", option=None):
        """tarball extract fail -> remove the tarball + remove .repo dir"""
        if option is None:
            option = []
        self.mySetupStep(tarball="/tarball." + suffix)
        self.expectClobber()
        self.expect_commands(
            self.ExpectShell(command=["tar", *option, "-xvf", "/tarball." + suffix]).exit(1),
            self.ExpectShell(command=['rm', '-f', '/tarball.tar']).exit(0),
            ExpectRmdir(dir='wkdir/.repo', log_environ=False).exit(0),
        )
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['stat', '-c%Y', '/tarball.' + suffix])
            .stdout(str(10000))
            .exit(0),
            self.ExpectShell(command=['stat', '-c%Y', '.'])
            .stdout(str(10001 + 7 * 24 * 3600))
            .exit(0),
            self.ExpectShell(command=["tar", *option, "-cvf", "/tarball." + suffix, ".repo"]).exit(
                0
            ),
        )
        return self.myRunStep()

    def test_update_tarball_fail2(self, suffix="tar", option=None):
        """tarball update fail -> remove the tarball + continue repo download"""
        if option is None:
            option = []
        self.mySetupStep(tarball="/tarball." + suffix)
        self.build.setProperty("repo_download", "repo download test/bla 564/12", "test")
        self.expectClobber()
        self.expect_commands(
            self.ExpectShell(command=["tar", *option, "-xvf", "/tarball." + suffix]).exit(0)
        )
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['stat', '-c%Y', '/tarball.' + suffix])
            .stdout(str(10000))
            .exit(0),
            self.ExpectShell(command=['stat', '-c%Y', '.'])
            .stdout(str(10001 + 7 * 24 * 3600))
            .exit(0),
            self.ExpectShell(command=["tar", *option, "-cvf", "/tarball." + suffix, ".repo"]).exit(
                1
            ),
            self.ExpectShell(command=['rm', '-f', '/tarball.tar']).exit(0),
            self.ExpectShell(command=['repo', 'download', 'test/bla', '564/12']).exit(0),
        )
        return self.myRunStep()

    def test_repo_downloads(self):
        """basic repo download, and check that repo_downloaded is updated"""
        self.mySetupStep()
        self.build.setProperty("repo_download", "repo download test/bla 564/12", "test")
        self.expectnoClobber()
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['repo', 'download', 'test/bla', '564/12'])
            .exit(0)
            .stderr("test/bla refs/changes/64/564/12 -> FETCH_HEAD\n")
            .stderr("HEAD is now at 0123456789abcdef...\n")
        )
        self.expect_property("repo_downloaded", "564/12 0123456789abcdef ", "Source")
        return self.myRunStep()

    def test_repo_downloads2(self):
        """2 repo downloads"""
        self.mySetupStep()
        self.build.setProperty("repo_download", "repo download test/bla 564/12", "test")
        self.build.setProperty("repo_download2", "repo download test/bla2 565/12", "test")
        self.expectnoClobber()
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['repo', 'download', 'test/bla', '564/12']).exit(0),
            self.ExpectShell(command=['repo', 'download', 'test/bla2', '565/12']).exit(0),
        )
        return self.myRunStep()

    def test_repo_download_manifest(self):
        """2 repo downloads, with one manifest patch"""
        self.mySetupStep()
        self.build.setProperty("repo_download", "repo download test/bla 564/12", "test")
        self.build.setProperty("repo_download2", "repo download manifest 565/12", "test")
        self.expectnoClobber()
        self.expect_commands(
            self.ExpectShell(
                command=['bash', '-c', self.get_nth_step(0)._getCleanupCommand()]
            ).exit(0),
            self.ExpectShell(
                command=[
                    'repo',
                    'init',
                    '-u',
                    'git://myrepo.com/manifest.git',
                    '-b',
                    'mb',
                    '-m',
                    'mf',
                    '--depth',
                    '0',
                ]
            ).exit(0),
            self.ExpectShell(
                workdir='wkdir/.repo/manifests',
                command=['git', 'fetch', 'git://myrepo.com/manifest.git', 'refs/changes/65/565/12'],
            ).exit(0),
            self.ExpectShell(
                workdir='wkdir/.repo/manifests', command=['git', 'cherry-pick', 'FETCH_HEAD']
            ).exit(0),
            self.ExpectShell(command=['repo', 'sync', '--force-sync', '-c']).exit(0),
            self.ExpectShell(
                command=['repo', 'manifest', '-r', '-o', 'manifest-original.xml']
            ).exit(0),
        )
        self.expect_commands(
            self.ExpectShell(command=['repo', 'download', 'test/bla', '564/12']).exit(0)
        )
        return self.myRunStep()

    def test_repo_downloads_mirror_sync(self):
        """repo downloads, with mirror synchronization issues"""
        self.mySetupStep()
        # we don't really want the test to wait...
        self.get_nth_step(0).mirror_sync_sleep = 0.001
        self.build.setProperty("repo_download", "repo download test/bla 564/12", "test")
        self.expectnoClobber()
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['repo', 'download', 'test/bla', '564/12'])
            .exit(1)
            .stderr("fatal: Couldn't find remote ref \n"),
            self.ExpectShell(command=['repo', 'download', 'test/bla', '564/12'])
            .exit(1)
            .stderr("fatal: Couldn't find remote ref \n"),
            self.ExpectShell(command=['repo', 'download', 'test/bla', '564/12']).exit(0),
        )
        return self.myRunStep()

    def test_repo_downloads_change_missing(self):
        """repo downloads, with no actual mirror synchronization issues (still retries 2 times)"""
        self.mySetupStep()
        # we don't really want the test to wait...
        self.get_nth_step(0).mirror_sync_sleep = 0.001
        self.get_nth_step(0).mirror_sync_retry = 1  # on retry once
        self.build.setProperty("repo_download", "repo download test/bla 564/12", "test")
        self.expectnoClobber()
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['repo', 'download', 'test/bla', '564/12'])
            .exit(1)
            .stderr("fatal: Couldn't find remote ref \n"),
            self.ExpectShell(command=['repo', 'download', 'test/bla', '564/12'])
            .exit(1)
            .stderr("fatal: Couldn't find remote ref \n"),
        )
        return self.myRunStep(
            result=FAILURE, state_string="repo: change test/bla 564/12 does not exist (failure)"
        )

    def test_repo_downloads_fail1(self):
        """repo downloads, cherry-pick returns 1"""
        self.mySetupStep()
        self.build.setProperty("repo_download", "repo download test/bla 564/12", "test")
        self.expectnoClobber()
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['repo', 'download', 'test/bla', '564/12'])
            .exit(1)
            .stderr("patch \n"),
            self.ExpectShell(command=['repo', 'forall', '-c', 'git', 'diff', 'HEAD']).exit(0),
        )
        return self.myRunStep(
            result=FAILURE, state_string="download failed: test/bla 564/12 (failure)"
        )

    def test_repo_downloads_fail2(self):
        """repo downloads, cherry-pick returns 0 but error in stderr"""
        self.mySetupStep()
        self.build.setProperty("repo_download", "repo download test/bla 564/12", "test")
        self.expectnoClobber()
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['repo', 'download', 'test/bla', '564/12'])
            .exit(0)
            .stderr("Automatic cherry-pick failed \n"),
            self.ExpectShell(command=['repo', 'forall', '-c', 'git', 'diff', 'HEAD']).exit(0),
        )
        return self.myRunStep(
            result=FAILURE, state_string="download failed: test/bla 564/12 (failure)"
        )

    def test_repo_downloads_from_change_source(self):
        """basic repo download from change source, and check that repo_downloaded is updated"""
        self.mySetupStep(repoDownloads=repo.RepoDownloadsFromChangeSource())
        change = Change(
            None,
            None,
            None,
            properties={
                'event.change.owner.email': 'dustin@mozilla.com',
                'event.change.subject': 'fix 1234',
                'event.change.project': 'pr',
                'event.change.owner.name': 'Dustin',
                'event.change.number': '4321',
                'event.change.url': 'http://buildbot.net',
                'event.change.branch': 'br',
                'event.type': 'patchset-created',
                'event.patchSet.revision': 'abcdef',
                'event.patchSet.number': '12',
                'event.source': 'GerritChangeSource',
            },
        )
        self.build.allChanges = lambda x=None: [change]
        self.expectnoClobber()
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['repo', 'download', 'pr', '4321/12'])
            .exit(0)
            .stderr("test/bla refs/changes/64/564/12 -> FETCH_HEAD\n")
            .stderr("HEAD is now at 0123456789abcdef...\n")
        )
        self.expect_property("repo_downloaded", "564/12 0123456789abcdef ", "Source")
        return self.myRunStep()

    def test_repo_downloads_from_change_source_codebase(self):
        """basic repo download from change source, and check that repo_downloaded is updated"""
        self.mySetupStep(repoDownloads=repo.RepoDownloadsFromChangeSource("mycodebase"))
        change = Change(
            None,
            None,
            None,
            properties={
                'event.change.owner.email': 'dustin@mozilla.com',
                'event.change.subject': 'fix 1234',
                'event.change.project': 'pr',
                'event.change.owner.name': 'Dustin',
                'event.change.number': '4321',
                'event.change.url': 'http://buildbot.net',
                'event.change.branch': 'br',
                'event.type': 'patchset-created',
                'event.patchSet.revision': 'abcdef',
                'event.patchSet.number': '12',
                'event.source': 'GerritChangeSource',
            },
        )
        # getSourceStamp is faked by SourceStepMixin
        ss = self.build.getSourceStamp("")
        ss.changes = [change]
        self.expectnoClobber()
        self.expectRepoSync()
        self.expect_commands(
            self.ExpectShell(command=['repo', 'download', 'pr', '4321/12'])
            .exit(0)
            .stderr("test/bla refs/changes/64/564/12 -> FETCH_HEAD\n")
            .stderr("HEAD is now at 0123456789abcdef...\n")
        )
        self.expect_property("repo_downloaded", "564/12 0123456789abcdef ", "Source")
        return self.myRunStep()

    def test_update_fail1(self):
        """fail at cleanup: ignored"""
        self.mySetupStep()
        self.expectnoClobber()
        self.expectRepoSync(which_fail=0, breakatfail=False)
        return self.myRunStep()

    def test_update_fail2(self):
        """fail at repo init: clobber"""
        self.mySetupStep()
        self.expectnoClobber()
        self.expectRepoSync(which_fail=1, breakatfail=True)
        self.expectClobber()
        self.expectRepoSync()
        self.shouldRetry = True
        return self.myRunStep()

    def test_update_fail3(self):
        """fail at repo sync: clobber"""
        self.mySetupStep()
        self.expectnoClobber()
        self.expectRepoSync(which_fail=2, breakatfail=True)
        self.expectClobber()
        self.expectRepoSync()
        self.shouldRetry = True
        return self.myRunStep()

    def test_update_fail4(self):
        """fail at repo manifest: clobber"""
        self.mySetupStep()
        self.expectnoClobber()
        self.expectRepoSync(which_fail=3, breakatfail=True)
        self.expectClobber()
        self.expectRepoSync()
        self.shouldRetry = True
        return self.myRunStep()

    def test_update_doublefail(self):
        """fail at repo manifest: clobber but still fail"""
        self.mySetupStep()
        self.expectnoClobber()
        self.expectRepoSync(which_fail=3, breakatfail=True)
        self.expectClobber()
        self.expectRepoSync(which_fail=3, breakatfail=True)
        self.shouldRetry = True
        return self.myRunStep(
            result=FAILURE, state_string="repo failed at: repo manifest (failure)"
        )

    def test_update_doublefail2(self):
        """fail at repo sync: clobber but still fail"""
        self.mySetupStep()
        self.expectnoClobber()
        self.expectRepoSync(which_fail=2, breakatfail=True)
        self.expectClobber()
        self.expectRepoSync(which_fail=2, breakatfail=True)
        self.shouldRetry = True
        return self.myRunStep(result=FAILURE, state_string="repo failed at: repo sync (failure)")

    def test_update_doublefail3(self):
        """fail at repo init: clobber but still fail"""
        self.mySetupStep()
        self.expectnoClobber()
        self.expectRepoSync(which_fail=1, breakatfail=True)
        self.expectClobber()
        self.expectRepoSync(which_fail=1, breakatfail=True)
        self.shouldRetry = True
        return self.myRunStep(result=FAILURE, state_string="repo failed at: repo init (failure)")

    def test_basic_fail(self):
        """fail at repo init: no need to re-clobber but still fail"""
        self.mySetupStep()
        self.expectClobber()
        self.expectRepoSync(which_fail=1, breakatfail=True)
        self.shouldRetry = True
        return self.myRunStep(result=FAILURE, state_string="repo failed at: repo init (failure)")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util import dirs
from buildbot.util import maildir


class TestMaildirService(dirs.DirsMixin, TestReactorMixin, unittest.TestCase):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.maildir = os.path.abspath("maildir")
        self.newdir = os.path.join(self.maildir, "new")
        self.curdir = os.path.join(self.maildir, "cur")
        self.tmpdir = os.path.join(self.maildir, "tmp")
        self.setUpDirs(self.maildir, self.newdir, self.curdir, self.tmpdir)

        self.master = yield fakemaster.make_master(self, wantDb=True, wantMq=True, wantData=True)

        self.svc = None

    def tearDown(self):
        if self.svc and self.svc.running:
            self.svc.stopService()

    # tests

    @defer.inlineCallbacks
    def test_start_stop_repeatedly(self):
        self.svc = maildir.MaildirService(self.maildir)
        yield self.svc.setServiceParent(self.master)
        yield self.master.startService()
        yield self.master.stopService()
        yield self.master.startService()
        yield self.master.stopService()
        self.assertEqual(len(list(self.svc)), 0)

    @defer.inlineCallbacks
    def test_messageReceived(self):
        self.svc = maildir.MaildirService(self.maildir)
        yield self.svc.setServiceParent(self.master)

        # add a fake messageReceived method
        messagesReceived = []

        def messageReceived(filename):
            messagesReceived.append(filename)
            return defer.succeed(None)

        self.svc.messageReceived = messageReceived
        yield self.master.startService()

        self.assertEqual(messagesReceived, [])

        tmpfile = os.path.join(self.tmpdir, "newmsg")
        newfile = os.path.join(self.newdir, "newmsg")
        with open(tmpfile, "w", encoding='utf-8'):
            pass
        os.rename(tmpfile, newfile)

        # TODO: can we wait for a dnotify somehow, if enabled?
        yield self.svc.poll()

        self.assertEqual(messagesReceived, ['newmsg'])

    def test_moveToCurDir(self):
        self.svc = maildir.MaildirService(self.maildir)
        tmpfile = os.path.join(self.tmpdir, "newmsg")
        newfile = os.path.join(self.newdir, "newmsg")
        with open(tmpfile, "w", encoding='utf-8'):
            pass
        os.rename(tmpfile, newfile)
        f = self.svc.moveToCurDir("newmsg")
        f.close()
        self.assertEqual(
            [
                os.path.exists(os.path.join(d, "newmsg"))
                for d in (self.newdir, self.curdir, self.tmpdir)
            ],
            [False, True, False],
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Portions Copyright Buildbot Team Members
# Portions Copyright 2007 Frederic Leroy <fredo@starox.org>
# Portions Copyright 2016 Louis Opter <kalessin@kalessin.fr>
#
#
# Documentation
# =============
#
# Mercurial "changegroup" hook that notifies Buildbot when a number of
# changsets is brought into the repository from elsewhere.
#
# Your Buildmaster needs to define a base ChangeHook, you should configure it
# behind a reverse proxy that does TLS and authentication for you and/or keep
# it behind a firewall. See the docs for more details:
#
# https://docs.buildbot.net/latest/manual/cfg-wwwhooks.html
#
# Copy this file to ".hg/hgbuildbot.py" in the repository that should notify
# Buildbot.
#
# Add it to the "[hooks]" section of ".hg/hgrc".  Also add a "[hgbuildbot]"
# section with additional parameters, for example:
#
#     [hooks]
#     changegroup.buildbot = python:.hg/hgbuiltbot.py:hook
#
#     [hgbuildbot]
#     venv = /home/buildbot/.virtualenvs/builtbot/lib/python2.7/site-packages
#     master = http://localhost:8020/change_hook/base
#
#
# Available parmeters
# -------------------
#
# venv
#   The hook needs the Python package "requests".  You can optionally point to
#   virtualenv if it is not installed globally:
#
#   Optional; default: None
#
#   Example:
#
#       venv = /path/to/venv/lib/pythonX.Y/site-packages
#
# master
#   URLs of the Buildmaster(s) to notify.
#   Can be a single entry or a comma-separated list.
#
#   Mandatory.
#
#   Examples:
#
#       master = localhost:8020/change_hook/base
#       master = bm1.example.org:8020/change_hook/base,bm2.example.org:8020/change_hook/base
#
# user
#   User for connecting to the Buildmaster. (Basic auth will be used).
#
#   Optional.
#
# passwd
#   Password for connecting to the Buildmaster.
#
#   Optional.
#
# branchtype
#   The branchmodel you use: "inrepo" for named branches (managed by
#   "hg branch") or "dirname" for directory based branches (the last component
#   of the repository's directory will then be used as branch name).
#
#   Optional; default: inrepo
#
# branch
#   Explicitly specify a branchname instead of using the repo's basename when
#   using "branchtype = dirname".
#
#   Optional.
#
# baseurl
#   Prefix for the repository URL sent to the Buildmaster.  See below for
#   details.
#
#   Optional.  The hook will also check the [web] section for this parameter.
#
# strip
#   Strip as many slashes from the repo dir before appending it to baseurl.
#   See below for details.
#
#   Optional; default: 0; The hook will also check the [notify] section for
#   this parameter.
#
# category
#   Category to assign to all change sets.
#
#   Optional.
#
# project
#   Project that the repo belongs to.
#
#   Optional.
#
# codebase
#   Codebase name for the repo.
#
#   Optional.
#
#
# Repository URLs
# ---------------
#
# The hook sends a repository URL to the Buildmasters.  It can be used by
# schedulers (e.g., for filtering) and is also used in the webview to create
# a link to the corresponding changeset.
#
# By default, the absolute repository path (e.g., "/home/hg/repos/myrepo") will
# be used.  The webview will in this case simply append the path to its own
# hostname in order to create a link to that change (e.g.,
# "http://localhost:8010/home/hg/repos/myrepo").
#
# You can alternatively strip some of the repo path's components and prepend
# a custom base URL instead.  For example, if you want to create an URL like
# "https://code.company.com/myrepo", you must specify the following parameters:
#
#     baseurl = https://code.company.com/
#     strip = 4
#
# This would strip everything until (and including) the 4th "/" in the repo's
# path leaving only "myrepo" left.  This would then be append to the base URL.

import json
import os
import os.path

import requests
from future.builtins import range
from mercurial.encoding import fromlocal
from mercurial.node import hex
from mercurial.node import nullid


def hook(ui, repo, hooktype, node=None, source=None, **kwargs):
    if hooktype != 'changegroup':
        ui.status(f'hgbuildbot: hooktype {hooktype} not supported.\n')
        return

    # Read config parameters
    masters = ui.configlist('hgbuildbot', 'master')
    if not masters:
        ui.write(
            '* You must add a [hgbuildbot] section to .hg/hgrc in order to use the Buildbot hook\n'
        )
        return

    # - virtualenv
    venv = ui.config('hgbuildbot', 'venv', None)
    if venv is not None:
        if not os.path.isdir(venv):
            ui.write(f'* Virtualenv "{venv}" does not exist.\n')
        else:
            activate_this = os.path.join(venv, "bin/activate_this.py")
            with open(activate_this) as f:
                activateThisScript = f.read()
            exec(activateThisScript, dict(__file__=activate_this))

    # - auth
    username = ui.config('hgbuildbot', 'user')
    password = ui.config('hgbuildbot', 'passwd')
    if username is not None and password is not None:
        auth = requests.auth.HTTPBasicAuth(username, password)
    else:
        auth = None

    # - branch
    branchtype = ui.config('hgbuildbot', 'branchtype', 'inrepo')
    branch = ui.config('hgbuildbot', 'branch', None)

    # - repo URL
    baseurl = ui.config('hgbuildbot', 'baseurl', ui.config('web', 'baseurl', ''))
    stripcount = int(ui.config('hgbuildbot', 'strip', ui.config('notify', 'strip', 0)))

    # - category, project and codebase
    category = ui.config('hgbuildbot', 'category', None)
    project = ui.config('hgbuildbot', 'project', '')
    codebase = ui.config('hgbuildbot', 'codebase', None)

    # Process changesets
    if branch is None and branchtype == 'dirname':
        branch = os.path.basename(repo.root)
    # If branchtype == 'inrepo', update "branch" for each commit later.

    repository = strip(repo.root, stripcount)
    repository = baseurl + repository

    start = repo[node].rev()
    end = len(repo)

    for rev in range(start, end):
        # send changeset
        node = repo.changelog.node(rev)
        log = repo.changelog.read(node)
        manifest, user, (time, timezone), files, desc, extra = log
        parents = [p for p in repo.changelog.parents(node) if p != nullid]

        if branchtype == 'inrepo':
            branch = extra['branch']
        if branch:
            branch = fromlocal(branch)

        is_merge = len(parents) > 1
        # merges don't always contain files, but at least one file is
        # required by buildbot
        if is_merge and not files:
            files = ["merge"]
        properties = {'is_merge': is_merge}

        change = {
            # 'master': master,
            'branch': branch,
            'revision': hex(node),
            'comments': fromlocal(desc),
            'files': json.dumps(files),
            'author': fromlocal(user),
            'category': category,
            'when': time,
            'properties': json.dumps(properties),
            'repository': repository,
            'project': project,
            'codebase': codebase,
        }
        for master in masters:
            response = requests.post(
                master,
                auth=auth,
                params=change,
                headers={"Content-Type": "application/x-www-form-urlencoded"},
            )
            if not response.ok:
                ui.warn(
                    f"couldn't notify buildbot about {hex(node)[:12]}: {response.status_code} {response.reason}"
                )
            else:
                ui.status(f"notified buildbot about {hex(node)[:12]}")


def strip(path, count):
    """Strip the count first slash of the path"""
    # First normalize it
    path = '/'.join(path.split(os.sep))
    # and strip the *count* first slash
    return path.split('/', count)[-1]



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

"""Interface documentation.

Define the interfaces that are implemented by various buildbot classes.
"""

# disable pylint warnings triggered by interface definitions
# pylint: disable=no-self-argument
# pylint: disable=no-method-argument
# pylint: disable=inherit-non-class

from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any

from zope.interface import Attribute
from zope.interface import Interface

if TYPE_CHECKING:
    from twisted.internet.defer import Deferred

    from buildbot.config.master import MasterConfig
    from buildbot.process.build import Build
    from buildbot.process.log import Log
    from buildbot.process.properties import Properties
    from buildbot.process.workerforbuilder import LatentWorkerForBuilder
    from buildbot.reporters.base import ReporterBase

# exceptions that can be raised while trying to start a build


class BuilderInUseError(Exception):
    pass


class WorkerSetupError(Exception):
    pass


class LatentWorkerFailedToSubstantiate(Exception):
    def __str__(self) -> str:
        return " ".join(str(arg) for arg in self.args)


class LatentWorkerCannotSubstantiate(Exception):
    def __str__(self) -> str:
        return " ".join(str(arg) for arg in self.args)


class LatentWorkerSubstantiatiationCancelled(Exception):
    def __str__(self) -> str:
        return " ".join(str(arg) for arg in self.args)


class IPlugin(Interface):
    """
    Base interface for all Buildbot plugins
    """


class IChangeSource(IPlugin):
    """
    Service which feeds Change objects to the changemaster. When files or
    directories are changed in version control, this object should represent
    the changes as a change dictionary and call::

      self.master.data.updates.addChange(who=.., rev=.., ..)

    See 'Writing Change Sources' in the manual for more information.
    """

    master = Attribute('master', 'Pointer to BuildMaster, automatically set when started.')

    def describe() -> str:
        """Return a string which briefly describes this source."""
        raise NotImplementedError


class ISourceStamp(Interface):
    """
    @cvar branch: branch from which source was drawn
    @type branch: string or None

    @cvar revision: revision of the source, or None to use CHANGES
    @type revision: varies depending on VC

    @cvar patch: patch applied to the source, or None if no patch
    @type patch: None or tuple (level diff)

    @cvar changes: the source step should check out the latest revision
                   in the given changes
    @type changes: tuple of L{buildbot.changes.changes.Change} instances,
                   all of which are on the same branch

    @cvar project: project this source code represents
    @type project: string

    @cvar repository: repository from which source was drawn
    @type repository: string
    """

    def canBeMergedWith(other: ISourceStamp) -> bool:
        """
        Can this SourceStamp be merged with OTHER?
        """
        raise NotImplementedError

    def mergeWith(others: list[ISourceStamp]) -> ISourceStamp:
        """Generate a SourceStamp for the merger of me and all the other
        SourceStamps. This is called by a Build when it starts, to figure
        out what its sourceStamp should be."""
        raise NotImplementedError

    def getAbsoluteSourceStamp(got_revision: str) -> ISourceStamp:
        """Get a new SourceStamp object reflecting the actual revision found
        by a Source step."""
        raise NotImplementedError

    def getText() -> str:
        """Returns a list of strings to describe the stamp. These are
        intended to be displayed in a narrow column. If more space is
        available, the caller should join them together with spaces before
        presenting them to the user."""
        raise NotImplementedError


class IEmailSender(Interface):
    """I know how to send email, and can be used by other parts of the
    Buildbot to contact developers."""


class IEmailLookup(Interface):
    def getAddress(user: str) -> Deferred:
        """Turn a User-name string into a valid email address. Either return
        a string (with an @ in it), None (to indicate that the user cannot
        be reached by email), or a Deferred which will fire with the same."""
        raise NotImplementedError


class ILogObserver(Interface):
    """Objects which provide this interface can be used in a BuildStep to
    watch the output of a LogFile and parse it incrementally.
    """

    # internal methods
    def setStep(step: IBuildStep) -> None:
        pass

    def setLog(log: Log) -> None:
        pass

    # methods called by the LogFile
    def logChunk(build: Build, step: IBuildStep, log: Log, channel: str, text: str) -> None:
        pass


class IWorker(IPlugin):
    # callback methods from the manager

    workername = Attribute('workername', 'The worker name')


class ILatentWorker(IWorker):
    """A worker that is not always running, but can run when requested."""

    substantiated = Attribute(
        'Substantiated',
        'Whether the latent worker is currently substantiated with a real instance.',
    )

    def substantiate(wfb: Any, build: Any) -> Deferred[Any]:
        """Request that the worker substantiate with a real instance.

        Returns a deferred that will callback when a real instance has
        attached."""
        raise NotImplementedError

    # there is an insubstantiate too, but that is not used externally ATM.

    def buildStarted(wfb: LatentWorkerForBuilder) -> None:
        """Inform the latent worker that a build has started.

        @param wfb: a L{LatentWorkerForBuilder}.  The wfb is the one for whom the
        build finished.
        """
        raise NotImplementedError

    def buildFinished(wfb: LatentWorkerForBuilder) -> None:
        """Inform the latent worker that a build has finished.

        @param wfb: a L{LatentWorkerForBuilder}.  The wfb is the one for whom the
        build finished.
        """
        raise NotImplementedError


class IMachine(Interface):
    pass


class IMachineAction(Interface):
    def perform(manager: IMachine) -> Deferred:
        """Perform an action on the machine managed by manager. Returns a
        deferred evaluating to True if it was possible to execute the
        action.
        """


class ILatentMachine(IMachine):
    """A machine that is not always running, but can be started when requested."""


class IRenderable(Interface):
    """An object that can be interpolated with properties from a build."""

    def getRenderingFor(iprops: IProperties) -> Deferred:
        """Return a deferred that fires with interpolation with the given properties

        @param iprops: the L{IProperties} provider supplying the properties.
        """
        raise NotImplementedError


class IProperties(Interface):
    """
    An object providing access to build properties
    """

    def getProperty(name: str, default: Any = None) -> object:
        """Get the named property, returning the default if the property does
        not exist.

        @param name: property name
        @type name: string

        @param default: default value (default: @code{None})

        @returns: property value
        """
        raise NotImplementedError

    def hasProperty(name: str) -> bool:
        """Return true if the named property exists.

        @param name: property name
        @type name: string
        @returns: boolean
        """
        raise NotImplementedError

    def has_key(name: str) -> bool:
        """Deprecated name for L{hasProperty}."""
        raise NotImplementedError

    def setProperty(name: str, value: object, source: str, runtime: bool = False) -> None:
        """Set the given property, overwriting any existing value.  The source
        describes the source of the value for human interpretation.

        @param name: property name
        @type name: string

        @param value: property value
        @type value: JSON-able value

        @param source: property source
        @type source: string

        @param runtime: (optional) whether this property was set during the
        build's runtime: usually left at its default value
        @type runtime: boolean
        """

    def getProperties() -> Properties:
        """Get the L{buildbot.process.properties.Properties} instance storing
        these properties.  Note that the interface for this class is not
        stable, so where possible the other methods of this interface should be
        used.

        @returns: L{buildbot.process.properties.Properties} instance
        """
        raise NotImplementedError

    def getBuild() -> Build:
        """Get the L{buildbot.process.build.Build} instance for the current
        build.  Note that this object is not available after the build is
        complete, at which point this method will return None.

        Try to avoid using this method, as the API of L{Build} instances is not
        well-defined.

        @returns L{buildbot.process.build.Build} instance
        """
        raise NotImplementedError

    def render(value: Any) -> Deferred[IRenderable]:
        """Render @code{value} as an L{IRenderable}.  This essentially coerces
        @code{value} to an L{IRenderable} and calls its @L{getRenderingFor}
        method.

        @name value: value to render
        @returns: rendered value
        """
        raise NotImplementedError


class IScheduler(IPlugin):
    pass


class ITriggerableScheduler(Interface):
    """
    A scheduler that can be triggered by buildsteps.
    """

    def trigger(
        waited_for, sourcestamps=None, set_props=None, parent_buildid=None, parent_relationship=None
    ):
        """Trigger a build with the given source stamp and properties."""


class IBuildStepFactory(Interface):
    def buildStep() -> IBuildStep:
        pass


class IBuildStep(IPlugin):
    """
    A build step
    """

    # Currently has nothing


class IConfigured(Interface):
    def getConfigDict() -> dict[str, Any]:
        return {}  # return something to silence warnings at call sites


class IReportGenerator(Interface):
    def generate(
        master: IConfigured, reporter: ReporterBase, key: str, build: Build
    ) -> Deferred[None]:
        raise NotImplementedError


class IConfigLoader(Interface):
    def loadConfig() -> MasterConfig:
        """
        Load the specified configuration.

        :return MasterConfig:
        """
        raise NotImplementedError


class IHttpResponse(Interface):
    def content() -> Deferred:
        """
        :returns: raw (``bytes``) content of the response via deferred
        """
        raise NotImplementedError

    def json() -> Deferred:
        """
        :returns: json decoded content of the response via deferred
        """
        raise NotImplementedError

    code = Attribute('code', "http status code of the request's response (e.g 200)")
    url = Attribute('url', "request's url (e.g https://api.github.com/endpoint')")


class IConfigurator(Interface):
    def configure(config_dict: dict[str, Any]) -> None:
        """
        Alter the buildbot config_dict, as defined in master.cfg

        like the master.cfg, this is run out of the main reactor thread, so this can block, but
        this can't call most Buildbot facilities.

        :returns: None
        """
        raise NotImplementedError
